"""
CSV validation utilities based on declarative rules.
"""

from __future__ import annotations

import csv
import json
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, List, Optional

import yaml


@dataclass
class CheckResult:
    name: str
    type: str
    status: str  # "pass" or "fail"
    message: str
    details: Dict[str, Any]


@dataclass
class ValidationSummary:
    checks_total: int
    checks_passed: int
    checks_failed: int
    row_count: int


@dataclass
class ValidationResult:
    summary: ValidationSummary
    checks: List[CheckResult]


def run_validation(csv_path: Path, rules_path: Path) -> ValidationResult:
    """
    Run validation checks defined in a YAML rules file against a CSV.
    """
    rows = _read_csv(csv_path)
    rules = _load_rules(rules_path)
    check_results: List[CheckResult] = []

    for rule in rules:
        result = _run_check(rule, rows)
        check_results.append(result)

    passed = sum(1 for r in check_results if r.status == "pass")
    failed = len(check_results) - passed
    summary = ValidationSummary(
        checks_total=len(check_results),
        checks_passed=passed,
        checks_failed=failed,
        row_count=len(rows),
    )
    return ValidationResult(summary=summary, checks=check_results)


def write_result_json(result: ValidationResult, path: Path) -> None:
    payload = {
        "summary": {
            "checks_total": result.summary.checks_total,
            "checks_passed": result.summary.checks_passed,
            "checks_failed": result.summary.checks_failed,
            "row_count": result.summary.row_count,
        },
        "checks": [
            {
                "name": c.name,
                "type": c.type,
                "status": c.status,
                "message": c.message,
                "details": c.details,
            }
            for c in result.checks
        ],
    }
    path.parent.mkdir(parents=True, exist_ok=True)
    path.write_text(json.dumps(payload, indent=2), encoding="utf-8")


def _read_csv(path: Path) -> List[Dict[str, str]]:
    with path.open("r", newline="", encoding="utf-8-sig") as handle:
        reader = csv.DictReader(handle)
        if reader.fieldnames is None:
            raise ValueError(f"CSV {path} is missing a header row.")
        return [dict(row) for row in reader]


def _load_rules(path: Path) -> List[Dict[str, Any]]:
    data = yaml.safe_load(path.read_text(encoding="utf-8"))
    if not isinstance(data, dict) or "checks" not in data:
        raise ValueError("Rules file must contain a top-level 'checks' list.")
    checks = data.get("checks") or []
    if not isinstance(checks, list):
        raise ValueError("'checks' must be a list.")
    normalized: List[Dict[str, Any]] = []
    for idx, entry in enumerate(checks):
        if not isinstance(entry, dict):
            raise ValueError(f"Check #{idx+1} must be a mapping.")
        if "name" not in entry or "type" not in entry:
            raise ValueError(f"Check #{idx+1} missing 'name' or 'type'.")
        normalized.append(entry)
    return normalized


def _run_check(rule: Dict[str, Any], rows: List[Dict[str, str]]) -> CheckResult:
    check_type = str(rule.get("type", "")).strip()
    name = str(rule.get("name", check_type))
    try:
        if check_type == "row_count_min":
            return _check_row_count_min(name, rows, rule)
        if check_type == "row_count_ratio":
            return _check_row_count_ratio(name, rows, rule)
        if check_type == "column_min":
            return _check_column_bound(name, rows, rule, bound_type="min")
        if check_type == "column_max":
            return _check_column_bound(name, rows, rule, bound_type="max")
        if check_type == "column_sum_range":
            return _check_column_sum_range(name, rows, rule)
        if check_type == "unique":
            return _check_unique(name, rows, rule)
        if check_type == "not_null":
            return _check_not_null(name, rows, rule)
        if check_type == "enum":
            return _check_enum(name, rows, rule)
        return _fail(name, check_type, f"Unsupported check type '{check_type}'.")
    except Exception as exc:
        return _fail(name, check_type, f"Error running check: {exc}")


def _check_row_count_min(name: str, rows: List[Dict[str, str]], rule: Dict[str, Any]) -> CheckResult:
    minimum = rule.get("min")
    if minimum is None:
        return _fail(name, "row_count_min", "Missing 'min' in rule.")
    row_count = len(rows)
    if row_count >= minimum:
        return _pass(name, "row_count_min", f"Row count {row_count} >= {minimum}.", {})
    return _fail(name, "row_count_min", f"Row count {row_count} < {minimum}.", {"row_count": row_count})


def _check_row_count_ratio(name: str, rows: List[Dict[str, str]], rule: Dict[str, Any]) -> CheckResult:
    baseline_path_raw = rule.get("baseline")
    min_ratio = rule.get("min_ratio")
    if baseline_path_raw is None or min_ratio is None:
        return _fail(name, "row_count_ratio", "Missing 'baseline' or 'min_ratio' in rule.")
    baseline_path = Path(baseline_path_raw)
    if not baseline_path.exists():
        return _fail(name, "row_count_ratio", f"Baseline file {baseline_path} not found.")
    baseline_rows = _read_csv(baseline_path)
    current = len(rows)
    baseline = len(baseline_rows)
    if baseline == 0:
        return _fail(name, "row_count_ratio", "Baseline row count is zero; cannot compute ratio.")
    ratio = current / baseline
    if ratio >= min_ratio:
        return _pass(
            name,
            "row_count_ratio",
            f"Row count ratio {ratio:.2f} >= {min_ratio}. (current={current}, baseline={baseline})",
            {"ratio": ratio, "current": current, "baseline": baseline},
        )
    return _fail(
        name,
        "row_count_ratio",
        f"Row count ratio {ratio:.2f} < {min_ratio}. (current={current}, baseline={baseline})",
        {"ratio": ratio, "current": current, "baseline": baseline},
    )


def _check_column_bound(name: str, rows: List[Dict[str, str]], rule: Dict[str, Any], bound_type: str) -> CheckResult:
    column = rule.get("column")
    bound_value = rule.get(bound_type)
    if not column or bound_value is None:
        return _fail(name, f"column_{bound_type}", f"Missing 'column' or '{bound_type}' in rule.")
    values = _coerce_numeric([row.get(column, "") for row in rows])
    if not values:
        return _fail(name, f"column_{bound_type}", f"No numeric values found in column '{column}'.")
    if bound_type == "min":
        violators = [val for val in values if val < bound_value]
        if not violators:
            return _pass(name, "column_min", f"All values in '{column}' >= {bound_value}.", {})
        return _fail(
            name,
            "column_min",
            f"{len(violators)} value(s) in '{column}' below {bound_value}.",
            {"violations": violators[:5]},
        )
    else:
        violators = [val for val in values if val > bound_value]
        if not violators:
            return _pass(name, "column_max", f"All values in '{column}' <= {bound_value}.", {})
        return _fail(
            name,
            "column_max",
            f"{len(violators)} value(s) in '{column}' above {bound_value}.",
            {"violations": violators[:5]},
        )


def _check_column_sum_range(name: str, rows: List[Dict[str, str]], rule: Dict[str, Any]) -> CheckResult:
    column = rule.get("column")
    min_sum = rule.get("min")
    max_sum = rule.get("max")
    if not column or min_sum is None or max_sum is None:
        return _fail(name, "column_sum_range", "Missing 'column', 'min', or 'max' in rule.")
    values = _coerce_numeric([row.get(column, "") for row in rows])
    total = sum(values)
    if total < min_sum or total > max_sum:
        return _fail(
            name,
            "column_sum_range",
            f"Sum of '{column}' = {total:.2f} outside [{min_sum}, {max_sum}].",
            {"sum": total},
        )
    return _pass(
        name,
        "column_sum_range",
        f"Sum of '{column}' = {total:.2f} within [{min_sum}, {max_sum}].",
        {"sum": total},
    )


def _check_unique(name: str, rows: List[Dict[str, str]], rule: Dict[str, Any]) -> CheckResult:
    column = rule.get("column")
    if not column:
        return _fail(name, "unique", "Missing 'column' in rule.")
    seen = {}
    duplicates = {}
    for idx, row in enumerate(rows, start=2):
        val = (row.get(column) or "").strip()
        if not val:
            continue
        if val in seen:
            duplicates.setdefault(val, []).append(idx)
        else:
            seen[val] = idx
    if not duplicates:
        return _pass(name, "unique", f"No duplicates in '{column}'.", {})
    sample = {k: v[:3] for k, v in list(duplicates.items())[:5]}
    return _fail(
        name,
        "unique",
        f"Found {len(duplicates)} duplicate value(s) in '{column}'.",
        {"duplicates": sample},
    )


def _check_not_null(name: str, rows: List[Dict[str, str]], rule: Dict[str, Any]) -> CheckResult:
    column = rule.get("column")
    if not column:
        return _fail(name, "not_null", "Missing 'column' in rule.")
    blanks = [idx for idx, row in enumerate(rows, start=2) if not (row.get(column) or "").strip()]
    if not blanks:
        return _pass(name, "not_null", f"All values present in '{column}'.", {})
    return _fail(
        name,
        "not_null",
        f"{len(blanks)} blank value(s) in '{column}'.",
        {"rows": blanks[:10]},
    )


def _check_enum(name: str, rows: List[Dict[str, str]], rule: Dict[str, Any]) -> CheckResult:
    column = rule.get("column")
    allowed = rule.get("allowed")
    if not column or not isinstance(allowed, list):
        return _fail(name, "enum", "Missing 'column' or 'allowed' list in rule.")
    allowed_set = {str(a) for a in allowed}
    invalid: Dict[str, List[int]] = {}
    for idx, row in enumerate(rows, start=2):
        raw = row.get(column)
        if raw is None:
            continue
        val = str(raw)
        if val not in allowed_set:
            invalid.setdefault(val, []).append(idx)
    if not invalid:
        return _pass(name, "enum", f"All values in '{column}' are allowed.", {})
    sample = {k: v[:3] for k, v in list(invalid.items())[:5]}
    return _fail(
        name,
        "enum",
        f"{len(invalid)} invalid value(s) in '{column}'.",
        {"invalid_values": sample},
    )


def _coerce_numeric(values: List[str]) -> List[float]:
    nums: List[float] = []
    for val in values:
        if val is None:
            continue
        text = str(val).replace(",", "").replace("$", "").strip()
        if not text:
            continue
        try:
            nums.append(float(text))
        except ValueError:
            continue
    return nums


def _pass(name: str, check_type: str, message: str, details: Dict[str, Any]) -> CheckResult:
    return CheckResult(name=name, type=check_type, status="pass", message=message, details=details)


def _fail(name: str, check_type: str, message: str, details: Optional[Dict[str, Any]] = None) -> CheckResult:
    return CheckResult(name=name, type=check_type, status="fail", message=message, details=details or {})



"""PandaDoc Public API client for direct document verification."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from g_gremlin.base import (
    AuthenticationError,
    BaseIntegration,
    IntegrationError,
    RateLimitError,
    RetryConfig,
)
from g_gremlin.credentials import resolve_credentials
from g_gremlin.credentials.provider import CredentialProvider

DEFAULT_API_BASE = "https://api.pandadoc.com/public/v1"
DEFAULT_RETRY_CONFIG = RetryConfig(max_attempts=4, base_delay=1.0, max_delay=10.0)


class PandaDocError(RuntimeError):
    """Raised when PandaDoc API access fails."""


@dataclass(frozen=True)
class PandaDocDocumentSnapshot:
    details: dict[str, Any]
    fields: list[dict[str, Any]]


def resolve_pandadoc_credentials(
    credential_provider: CredentialProvider | None = None,
) -> dict[str, str]:
    provider = credential_provider or CredentialProvider()
    return resolve_credentials(
        "pandadoc",
        required_keys=["api_key"],
        optional_keys=["api_base"],
        credential_provider=provider,
    )


class PandaDocClient(BaseIntegration):
    SERVICE_NAME = "pandadoc"

    def __init__(
        self,
        *,
        api_key: str | None = None,
        api_base: str | None = None,
        credential_provider: CredentialProvider | None = None,
        retry_config: RetryConfig | None = None,
    ) -> None:
        super().__init__(
            credential_provider=credential_provider,
            default_retry_config=retry_config or DEFAULT_RETRY_CONFIG,
        )
        if api_key is None:
            creds = resolve_pandadoc_credentials(credential_provider)
            api_key = creds["api_key"]
            api_base = api_base or creds.get("api_base")
        self.api_key = str(api_key or "").strip()
        self.api_base = _normalize_api_base(api_base)

    def get_document_snapshot(self, document_id: str) -> PandaDocDocumentSnapshot:
        normalized_id = str(document_id or "").strip()
        if not normalized_id:
            raise PandaDocError("PandaDoc document ID is required.")
        details = self.get_document_details(normalized_id)
        fields = self.list_document_fields(normalized_id)
        return PandaDocDocumentSnapshot(details=details, fields=fields)

    def get_document_details(self, document_id: str) -> dict[str, Any]:
        payload = self._get_json(
            f"{self.api_base}/documents/{document_id}/details",
            allow_statuses=(),
        )
        if not isinstance(payload, dict):
            raise PandaDocError("PandaDoc document details payload was not an object.")
        return payload

    def list_document_fields(self, document_id: str) -> list[dict[str, Any]]:
        payload = self._get_json(
            f"{self.api_base}/documents/{document_id}/fields",
            allow_statuses=(404,),
        )
        if not payload:
            return []
        if isinstance(payload, list):
            return [item for item in payload if isinstance(item, dict)]
        if not isinstance(payload, dict):
            return []
        for key in ("results", "fields", "data"):
            raw_items = payload.get(key)
            if isinstance(raw_items, list):
                return [item for item in raw_items if isinstance(item, dict)]
        return []

    def _headers(self) -> dict[str, str]:
        return {
            "Authorization": f"API-Key {self.api_key}",
            "Accept": "application/json",
        }

    def _get_json(
        self,
        url: str,
        *,
        allow_statuses: tuple[int, ...],
    ) -> dict[str, Any] | list[Any]:
        try:
            response = self._get(
                url,
                headers=self._headers(),
                allow_statuses=allow_statuses,
            )
        except (AuthenticationError, RateLimitError, IntegrationError) as exc:
            raise PandaDocError(str(exc)) from exc
        if response.status_code in allow_statuses:
            return {}
        try:
            payload = response.json()
        except Exception as exc:
            raise PandaDocError(f"Failed to parse PandaDoc response from {url}: {exc}") from exc
        if isinstance(payload, (dict, list)):
            return payload
        raise PandaDocError(f"Unexpected PandaDoc response payload from {url}.")


def _normalize_api_base(value: Any) -> str:
    normalized = str(value or "").strip().rstrip("/")
    if not normalized:
        return DEFAULT_API_BASE
    return normalized


__all__ = [
    "DEFAULT_API_BASE",
    "PandaDocClient",
    "PandaDocDocumentSnapshot",
    "PandaDocError",
    "resolve_pandadoc_credentials",
]



"""
WebhookEmitter - Sends AgenticResult payloads to webhooks.

Handles profile resolution, size limits, format transformation, and retry logic.
"""

from __future__ import annotations

import json
from dataclasses import dataclass
from typing import Any, Dict, Optional

from urllib.parse import urlparse

from g_gremlin.contracts import AgenticResult
from g_gremlin.retry_utils import RetryConfig, _compute_delay, _sleep
from g_gremlin.webhook_out.envelope import WebhookEnvelope
from g_gremlin.webhook_out.profiles import profile_get
from g_gremlin.webhook_out.transformers import get_transformer


# Maximum payload size before auto-truncation (1.5MB - safe for all iPaaS)
MAX_PAYLOAD_BYTES = 1_500_000

# Allowed URL schemes to prevent SSRF attacks
ALLOWED_SCHEMES = {"https", "http"}


@dataclass
class WebhookDeliveryResult:
    """Result of webhook delivery attempt."""

    success: bool
    delivery_id: str
    status_code: Optional[int] = None
    error: Optional[str] = None
    attempts: int = 1

    def to_dict(self) -> Dict[str, Any]:
        return {
            "success": self.success,
            "delivery_id": self.delivery_id,
            "status_code": self.status_code,
            "error": self.error,
            "attempts": self.attempts,
        }


def _maybe_truncate(
    envelope: WebhookEnvelope,
    force_summary: bool = False,
    force_full: bool = False,
) -> WebhookEnvelope:
    """
    Truncate payload if too large, always preserving envelope fields.

    Args:
        envelope: The envelope to potentially truncate
        force_summary: If True, always truncate to summary
        force_full: If True, never truncate (may exceed size limits)

    Returns:
        Original envelope or truncated version
    """
    if force_full:
        return envelope

    # Serialize once for size check
    serialized = json.dumps(envelope.to_dict()).encode("utf-8")

    if force_summary or len(serialized) > MAX_PAYLOAD_BYTES:
        # Extract summary, preserve all envelope fields
        payload = envelope.payload
        truncated_payload = {
            "$truncated": True,
            "original_bytes": len(serialized),
            "summary": {
                "$schema": payload.get("$schema"),
                "status": payload.get("status"),
                "command": payload.get("command"),
                "correlation_id": payload.get("correlation_id"),
                "dry_run": payload.get("dry_run"),
                "error": payload.get("error"),
                "timestamp": payload.get("timestamp"),
            },
            "message": "Payload exceeded 1.5MB. Use --output to save full result locally.",
        }

        return WebhookEnvelope(
            event_type=envelope.event_type,
            correlation_id=envelope.correlation_id,
            delivery_id=envelope.delivery_id,
            source=envelope.source,
            source_version=envelope.source_version,
            timestamp=envelope.timestamp,
            payload=truncated_payload,
            attempt=envelope.attempt,
            max_attempts=envelope.max_attempts,
        )

    return envelope


class WebhookEmitter:
    """
    Emits AgenticResult payloads to webhook endpoints.

    Supports:
    - Direct URL or @profile-name references
    - Format-specific transformations (flat, wrapped, generic)
    - Configurable retry with exponential backoff
    - Automatic payload truncation for large results
    """

    DEFAULT_TIMEOUT = 10.0  # Fast timeout for webhooks
    DEFAULT_RETRY_CONFIG = RetryConfig(
        max_attempts=3,
        base_delay=1.0,
        max_delay=30.0,
        exponential_base=2.0,
        jitter=True,
        retryable_codes={429, 500, 502, 503, 504},
    )

    def __init__(
        self,
        timeout: Optional[float] = None,
        retry_config: Optional[RetryConfig] = None,
    ):
        self.timeout = timeout or self.DEFAULT_TIMEOUT
        self.retry_config = retry_config or self.DEFAULT_RETRY_CONFIG
        self._httpx = None

    def _get_httpx(self):
        """Lazy import httpx."""
        if self._httpx is None:
            import httpx

            self._httpx = httpx
        return self._httpx

    def _validate_url(self, url: str) -> None:
        """
        Validate URL scheme to prevent SSRF attacks.

        Only https:// and http:// are allowed.
        Rejects file://, ftp://, gopher://, localhost without scheme, etc.
        """
        try:
            parsed = urlparse(url)
            if parsed.scheme.lower() not in ALLOWED_SCHEMES:
                raise ValueError(
                    f"Invalid URL scheme '{parsed.scheme}'. Only https:// and http:// are allowed."
                )
            if not parsed.netloc:
                raise ValueError("Invalid URL: missing host")
        except Exception as e:
            if "Invalid URL" in str(e) or "scheme" in str(e):
                raise
            raise ValueError(f"Invalid URL: {e}")

    def resolve_target(
        self,
        webhook_out: str,
        format_override: Optional[str] = None,
    ) -> tuple[str, Optional[str], Dict[str, str], float, int]:
        """
        Resolve webhook target from URL or @profile reference.

        Returns:
            (url, format, headers, timeout, max_retries)

        Raises:
            ValueError: If profile not found or URL scheme is invalid
        """
        if webhook_out.startswith("@"):
            profile_name = webhook_out[1:]  # Strip @
            profile = profile_get(profile_name)
            if profile is None:
                raise ValueError(f"Webhook profile '{profile_name}' not found")
            # Validate URL from profile
            self._validate_url(profile.url)
            return (
                profile.url,
                format_override or profile.format,
                profile.headers,
                profile.timeout,
                profile.max_retries,
            )

        # Direct URL - validate and use defaults
        self._validate_url(webhook_out)
        return (
            webhook_out,
            format_override,
            {},
            self.timeout,
            self.retry_config.max_attempts,
        )

    def emit(
        self,
        result: AgenticResult,
        webhook_out: str,
        format: Optional[str] = None,
        force_summary: bool = False,
        force_full: bool = False,
        extra_headers: Optional[Dict[str, str]] = None,
    ) -> WebhookDeliveryResult:
        """
        Send AgenticResult to webhook endpoint.

        Args:
            result: The AgenticResult to send
            webhook_out: URL or @profile-name
            format: Format preset (flat|wrapped|generic), overrides profile
            force_summary: Force summary-only payload
            force_full: Force full payload (may exceed size limits)
            extra_headers: Additional headers to include

        Returns:
            WebhookDeliveryResult with delivery status
        """
        httpx = self._get_httpx()

        # Resolve target
        url, resolved_format, profile_headers, timeout, max_retries = self.resolve_target(
            webhook_out, format
        )

        # Build envelope from result
        envelope = WebhookEnvelope.from_agentic_result(result, max_attempts=max_retries)

        # Apply size limits
        envelope = _maybe_truncate(envelope, force_summary=force_summary, force_full=force_full)

        # Transform for format
        transformer = get_transformer(resolved_format)
        body, transform_headers = transformer(envelope)

        # Merge headers (profile < transform < extra)
        final_headers = {**profile_headers, **transform_headers}
        if extra_headers:
            final_headers.update(extra_headers)

        # Update retry config with profile settings
        retry_config = RetryConfig(
            max_attempts=max_retries,
            base_delay=self.retry_config.base_delay,
            max_delay=self.retry_config.max_delay,
            exponential_base=self.retry_config.exponential_base,
            jitter=self.retry_config.jitter,
            retryable_codes=self.retry_config.retryable_codes,
        )

        # Retry loop
        last_error: Optional[str] = None
        last_status: Optional[int] = None

        for attempt in range(1, retry_config.max_attempts + 1):
            envelope.attempt = attempt

            try:
                resp = httpx.post(
                    url,
                    json=body,
                    headers=final_headers,
                    timeout=timeout,
                )
                last_status = resp.status_code

                if 200 <= resp.status_code < 300:
                    return WebhookDeliveryResult(
                        success=True,
                        delivery_id=envelope.delivery_id,
                        status_code=resp.status_code,
                        attempts=attempt,
                    )

                # Check if retryable
                if resp.status_code in retry_config.retryable_codes:
                    last_error = f"HTTP {resp.status_code}"
                    if attempt < retry_config.max_attempts:
                        delay = _compute_delay(retry_config, attempt)
                        _sleep(delay)
                        continue

                # Non-retryable error
                return WebhookDeliveryResult(
                    success=False,
                    delivery_id=envelope.delivery_id,
                    status_code=resp.status_code,
                    error=f"HTTP {resp.status_code}",
                    attempts=attempt,
                )

            except Exception as exc:
                last_error = str(exc)
                if attempt < retry_config.max_attempts:
                    delay = _compute_delay(retry_config, attempt)
                    _sleep(delay)
                    continue

        # All retries exhausted
        return WebhookDeliveryResult(
            success=False,
            delivery_id=envelope.delivery_id,
            status_code=last_status,
            error=last_error or "Max retries exceeded",
            attempts=retry_config.max_attempts,
        )



import logging
import httpx
from sqlalchemy import select

from .. import settings
from ..db import AsyncSessionLocal

log = logging.getLogger(__name__)


class SlackActor:
    def __init__(self, slack_user_id: str, user_id: str, account_id: str, role: str):
        self.slack_user_id = slack_user_id
        self.user_id = user_id
        self.account_id = account_id
        self.role = role


async def map_slack_user(slack_user_id: str) -> SlackActor | None:
    try:
        # Dev shortcut
        if (settings.ENV or "").lower().startswith("dev"):
            return SlackActor(
                slack_user_id, "dev-user", settings.DEV_ACCOUNT_ID, "admin"
            )

        token = settings.SLACK_BOT_TOKEN
        if not token:
            return None
        async with httpx.AsyncClient(timeout=10) as client:
            r = await client.get(
                "https://slack.com/api/users.info",
                params={"user": slack_user_id},
                headers={"Authorization": f"Bearer {token}"},
            )
            data = r.json()
        if not data.get("ok"):
            return None
        email = (
            (((data.get("user") or {}).get("profile") or {}).get("email") or "")
            .lower()
            .strip()
        )
        if not email:
            return None
        from ...models import User

        async with AsyncSessionLocal() as db:
            row = (
                await db.execute(select(User).where(User.email == email))
            ).scalar_one_or_none()
            if not row:
                return None
            role = getattr(row, "role", "viewer")
            role = getattr(role, "value", role)
            return SlackActor(
                slack_user_id, str(row.id), str(row.account_id or ""), str(role)
            )
    except Exception:
        log.exception("map_slack_user failed")
        return None



# In tests/test_mapping.py

# Corrected import - we remove the import that is causing the failure.
# The internal logic of your tests here will need to be updated to reflect
# the new engine structure, likely by calling the AutoMapper class.

def test_mapping_functionality():
    # Placeholder for your mapping test logic.
    # The primary goal is to fix the import error so pytest can run.
    assert True



"""
Feedback API for Match Quality Learning
========================================
Collects user feedback on match quality to improve future configurations.
"""

from fastapi import APIRouter, Depends, HTTPException
from typing import Dict, Any, Optional
from uuid import UUID
from pydantic import BaseModel
import logging

from ...auth import get_current_account
from ...db import get_session
from sqlalchemy.ext.asyncio import AsyncSession
from ...intelligence.decision_logger import DecisionLogger
from ...intelligence.bandits import ThresholdBandit, BlockingBandit
import redis.asyncio as async_redis

log = logging.getLogger(__name__)
router = APIRouter(prefix="/api/v2/feedback", tags=["feedback"])


# Pydantic models
class MatchFeedback(BaseModel):
    match_id: str
    job_id: str
    outcome: str  # 'accepted', 'reverted', 'corrected'
    corrected_target: Optional[str] = None
    confidence: Optional[float] = None  # User's confidence in their feedback
    notes: Optional[str] = None


class BulkFeedback(BaseModel):
    job_id: str
    accepted_matches: list[str] = []
    reverted_matches: list[str] = []
    corrected_matches: list[
        Dict[str, str]
    ] = []  # [{"match_id": "...", "correct_target": "..."}]


# Get Redis connection
async def get_redis():
    """Get Redis connection for feedback storage"""
    try:
        redis_conn = async_redis.Redis(
            host="localhost", port=6379, decode_responses=False
        )
        await redis_conn.ping()
        return redis_conn
    except Exception as e:
        log.error(f"Redis connection failed: {e}")
        return None


@router.post("/match")
async def record_match_feedback(
    feedback: MatchFeedback,
    db: AsyncSession = Depends(get_session),
    account_id: UUID = Depends(get_current_account),
) -> Dict[str, Any]:
    """
    Record feedback on a single match.
    Updates bandit statistics for continuous learning.
    """
    redis_conn = await get_redis()
    logger = DecisionLogger(redis_conn=redis_conn)

    # Log the feedback
    await logger.log_feedback(
        job_id=feedback.job_id,
        match_id=feedback.match_id,
        outcome=feedback.outcome,
        user_id=str(account_id),
        corrected_target=feedback.corrected_target,
    )

    # DecisionLogger will handle bandit updates internally

    return {
        "status": "recorded",
        "match_id": feedback.match_id,
        "message": f"Feedback recorded as '{feedback.outcome}'",
    }


@router.post("/bulk")
async def record_bulk_feedback(
    feedback: BulkFeedback,
    db: AsyncSession = Depends(get_session),
    account_id: UUID = Depends(get_current_account),
) -> Dict[str, Any]:
    """
    Record feedback on multiple matches at once.
    Useful for batch review workflows.
    """
    redis_conn = await get_redis()
    logger = DecisionLogger(redis_conn=redis_conn)

    processed = 0

    # Process accepted matches
    for match_id in feedback.accepted_matches:
        await logger.log_feedback(
            job_id=feedback.job_id,
            match_id=match_id,
            outcome="accepted",
            user_id=str(account_id),
        )
        processed += 1

    # Process reverted matches
    for match_id in feedback.reverted_matches:
        await logger.log_feedback(
            job_id=feedback.job_id,
            match_id=match_id,
            outcome="reverted",
            user_id=str(account_id),
        )
        processed += 1

    # Process corrected matches
    for correction in feedback.corrected_matches:
        await logger.log_feedback(
            job_id=feedback.job_id,
            match_id=correction["match_id"],
            outcome="corrected",
            user_id=str(account_id),
            corrected_target=correction.get("correct_target"),
        )
        processed += 1

    # Trigger batch learning update
    if processed > 0 and redis_conn:
        await _trigger_batch_learning(feedback.job_id, feedback, redis_conn)

    return {
        "status": "recorded",
        "job_id": feedback.job_id,
        "processed_count": processed,
        "breakdown": {
            "accepted": len(feedback.accepted_matches),
            "reverted": len(feedback.reverted_matches),
            "corrected": len(feedback.corrected_matches),
        },
    }


@router.get("/stats/{job_id}")
async def get_feedback_stats(
    job_id: str,
    db: AsyncSession = Depends(get_session),
    account_id: UUID = Depends(get_current_account),
) -> Dict[str, Any]:
    """
    Get feedback statistics for a job.
    Shows learning progress and configuration performance.
    """
    redis_conn = await get_redis()

    if not redis_conn:
        raise HTTPException(503, "Statistics service unavailable")

    import json

    # Get decision data
    decision_data = await redis_conn.get(f"decision:{job_id}")
    if not decision_data:
        raise HTTPException(404, f"No decision data found for job {job_id}")

    decision = json.loads(decision_data)
    segment = decision.get("segment", "unknown")

    # Get bandit performance
    threshold_bandit = ThresholdBandit(redis_conn)
    blocking_bandit = BlockingBandit(redis_conn)

    threshold_stats = await threshold_bandit.get_performance_summary(segment)
    blocking_stats = (
        await blocking_bandit.get_performance_summary(segment)
        if decision.get("blocking")
        else None
    )

    # Get feedback counts from logger
    logger = DecisionLogger(redis_conn=redis_conn)
    segment_perf = await logger.get_segment_performance(segment, days=30)

    return {
        "job_id": job_id,
        "segment": segment,
        "configuration_used": {
            "threshold_policy": decision.get("threshold"),
            "blocking_strategy": decision.get("blocking"),
        },
        "learning_stats": {"threshold": threshold_stats, "blocking": blocking_stats},
        "segment_performance": segment_perf,
        "message": "Learning system is adapting based on feedback",
    }


async def _trigger_batch_learning(job_id: str, feedback: BulkFeedback, redis_conn):
    """
    Process bulk feedback for learning.
    Calculates aggregate reward and updates bandits.
    """
    import json

    # Calculate aggregate reward
    total = (
        len(feedback.accepted_matches)
        + len(feedback.reverted_matches)
        + len(feedback.corrected_matches)
    )
    if total == 0:
        return

    accepted_weight = len(feedback.accepted_matches) / total
    reverted_weight = len(feedback.reverted_matches) / total
    corrected_weight = len(feedback.corrected_matches) / total

    # Weighted average reward
    aggregate_reward = (
        accepted_weight * 1.0 + corrected_weight * 0.3 + reverted_weight * 0.0
    )

    # Get decision configuration
    decision_data = await redis_conn.get(f"decision:{job_id}")
    if decision_data:
        decision = json.loads(decision_data)
        segment = decision.get("segment", "unknown")

        # Update bandits with aggregate reward
        if decision.get("threshold"):
            threshold_bandit = ThresholdBandit(redis_conn)
            await threshold_bandit.update(
                segment, decision["threshold"], aggregate_reward
            )

        if decision.get("blocking"):
            blocking_bandit = BlockingBandit(redis_conn)
            await blocking_bandit.update(
                segment, decision["blocking"], aggregate_reward
            )

        log.info(
            f"Batch learning update for job {job_id}: "
            f"aggregate_reward={aggregate_reward:.3f} from {total} feedback items"
        )



from __future__ import annotations
from fastapi import APIRouter, Depends, HTTPException, Header, Query, Body
from sqlalchemy.orm import Session
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm.exc import StaleDataError
from typing import Optional, List
from ..db import get_session
from ..auth_core import require_admin, get_current_account
from .schemas import (
    RouteRequest,
    RouteResponse,
    PoolCreate,
    PoolMemberUpsert,
    RuleUpsert,
    LeadIn,
)
from .models import (
    RoutingPool,
    RoutingPoolMember,
    RoutingRule,
    RoutingDecision,
    RoutingPoolState,
)
from .engine import route_lead
from .adapters.salesforce import SalesforceAdapter
from ..services.salesforce_gateway import SalesforceGateway
from .slack import post_decision
from datetime import datetime, timedelta
from zoneinfo import ZoneInfo
import logging
import json
import uuid
import hashlib
import time
from ..models import Integration, IntegrationProvider
from sqlalchemy import select

router = APIRouter(prefix="/api/v2/routing", tags=["routing"])
admin_router = APIRouter(
    prefix="/api/v2/routing/admin",
    tags=["routing-admin"],
    dependencies=[Depends(require_admin)],
)
logger = logging.getLogger(__name__)


async def get_salesforce_gateway_for_tenant(
    db: Session, tenant_id: str
) -> Optional[SalesforceGateway]:
    """Get Salesforce gateway for the given tenant"""
    # Ensure an integration exists
    result = await db.execute(
        select(Integration).where(
            Integration.account_id == tenant_id,
            Integration.provider == IntegrationProvider.SALESFORCE,
        )
    )
    integration = result.scalar_one_or_none()
    if not integration:
        return None
    return await SalesforceGateway.for_tenant(tenant_id, db)


def validate_tenant_access(tenant_id: str, current_account: dict) -> bool:
    """Validate that the current account has access to the specified tenant"""
    # For multi-tenant systems, validate that the account belongs to the tenant
    # This is a simplified check - in production you'd validate against account permissions
    account_tenant = current_account.get("tenant_id") or current_account.get("id")
    if not account_tenant:
        return False
    # Allow access if account tenant matches or if account is a super admin
    return account_tenant == tenant_id or current_account.get("is_super_admin", False)


@router.post("/route/preview", response_model=RouteResponse)
async def preview_route(
    payload: RouteRequest,
    db: Session = Depends(get_session),
    current_account: dict = Depends(get_current_account),
) -> RouteResponse:
    # Validate tenant access
    if not validate_tenant_access(payload.tenant_id, current_account):
        raise HTTPException(403, "Access denied to this tenant")

    # Basic input validation (fast fail)
    if not (payload.lead.email or payload.lead.company or payload.lead.website):
        raise HTTPException(400, "Lead must include email, company, or website")
    sf = await get_salesforce_gateway_for_tenant(db, tenant_id=payload.tenant_id)
    adapter = SalesforceAdapter(sf) if sf else SalesforceAdapter(None)
    decision = await route_lead(
        db, payload.lead, payload.tenant_id, adapter, preview=True
    )
    # metrics/log
    logger.info(
        "route_preview tenant=%s decision=%s context=%s",
        payload.tenant_id,
        decision.get("decision"),
        json.dumps(decision.get("context") or {})[:500],
    )
    return RouteResponse(**decision)


@router.post("/route", response_model=RouteResponse)
async def commit_route(
    payload: RouteRequest,
    db: Session = Depends(get_session),
    current_account: dict = Depends(get_current_account),
    x_idempotency_key: Optional[str] = Header(None),
) -> RouteResponse:
    # Validate tenant access
    if not validate_tenant_access(payload.tenant_id, current_account):
        raise HTTPException(403, "Access denied to this tenant")

    if not (payload.lead.email or payload.lead.company or payload.lead.website):
        raise HTTPException(400, "Lead must include email, company, or website")

    # Check for idempotency key in header or payload
    idempotency_key = x_idempotency_key or getattr(payload, "idempotency_key", None)

    # If idempotency key provided, check for existing decision
    if idempotency_key:
        existing = (
            db.query(RoutingDecision)
            .filter_by(tenant_id=payload.tenant_id, idempotency_key=idempotency_key)
            .first()
        )

        if existing:
            # Return existing decision
            logger.info(
                "route_commit_idempotent tenant=%s idempotency_key=%s decision_id=%s",
                payload.tenant_id,
                idempotency_key,
                existing.id,
            )
            return RouteResponse(
                decision=existing.decision,
                owner_id=existing.owner_id,
                pool_id=existing.pool_id,
                rule_id=existing.rule_id,
                reason=existing.reason,
                context=existing.context or {},
            )

    # Generate idempotency key if not provided (for deduplication)
    if not idempotency_key:
        # Create hash from lead data for automatic dedup
        lead_data = f"{payload.tenant_id}:{payload.lead.email}:{payload.lead.company}:{payload.lead.website}"
        idempotency_key = hashlib.sha256(lead_data.encode()).hexdigest()[:32]

    sf = await get_salesforce_gateway_for_tenant(db, tenant_id=payload.tenant_id)
    adapter = SalesforceAdapter(sf) if sf else SalesforceAdapter(None)
    decision = await route_lead(
        db, payload.lead, payload.tenant_id, adapter, preview=False
    )

    # persist audit with idempotency key
    rd = RoutingDecision(
        id=str(uuid.uuid4()),
        tenant_id=payload.tenant_id,
        lead_id=payload.lead.lead_id or "unknown",
        decision=decision["decision"],
        owner_id=decision.get("owner_id"),
        pool_id=decision.get("pool_id"),
        rule_id=decision.get("rule_id"),
        reason=decision.get("reason"),
        context=decision.get("context") or {},
        created_at=datetime.utcnow(),
        applied_at=datetime.utcnow(),  # Mark as applied
        idempotency_key=idempotency_key,
    )

    try:
        db.add(rd)
        db.commit()
    except IntegrityError:
        # Race condition - another request with same idempotency key completed
        db.rollback()
        existing = (
            db.query(RoutingDecision)
            .filter_by(tenant_id=payload.tenant_id, idempotency_key=idempotency_key)
            .first()
        )
        if existing:
            return RouteResponse(
                decision=existing.decision,
                owner_id=existing.owner_id,
                pool_id=existing.pool_id,
                rule_id=existing.rule_id,
                reason=existing.reason,
                context=existing.context or {},
            )
        raise

    # optional notify
    await post_decision(decision)
    logger.info(
        "route_commit tenant=%s decision=%s idempotency_key=%s context=%s",
        payload.tenant_id,
        decision.get("decision"),
        idempotency_key,
        json.dumps(decision.get("context") or {})[:500],
    )
    return RouteResponse(**decision)


# ----------------------- Admin Endpoints -----------------------


@admin_router.get("/pools")
def list_pools(
    tenant_id: str,
    db: Session = Depends(get_session),
    current_account: dict = Depends(get_current_account),
):
    # Validate tenant access
    if not validate_tenant_access(tenant_id, current_account):
        raise HTTPException(403, "Access denied to this tenant")
    pools = db.query(RoutingPool).filter_by(tenant_id=tenant_id).all()
    members = db.query(RoutingPoolMember).filter_by(tenant_id=tenant_id).all()
    by_pool = {}
    for m in members:
        by_pool.setdefault(m.pool_id, []).append(
            {"user_id": m.user_id, "weight": m.weight}
        )
    return [
        {
            "id": p.id,
            "name": p.name,
            "algo": p.algo,
            "daily_cap": p.daily_cap,
            "members": by_pool.get(p.id, []),
        }
        for p in pools
    ]


@admin_router.post("/pools")
def upsert_pool(
    body: PoolCreate,
    db: Session = Depends(get_session),
    current_account: dict = Depends(get_current_account),
):
    # Validate tenant access
    if not validate_tenant_access(body.tenant_id, current_account):
        raise HTTPException(403, "Access denied to this tenant")
    p = (
        db.query(RoutingPool)
        .filter_by(tenant_id=body.tenant_id, id=body.pool_id)
        .one_or_none()
    )
    if not p:
        p = RoutingPool(
            id=body.pool_id,
            tenant_id=body.tenant_id,
            name=body.name,
            algo=body.algo,
            daily_cap=body.daily_cap,
        )
        db.add(p)
    else:
        p.name, p.algo, p.daily_cap = body.name, body.algo, body.daily_cap
    db.commit()
    return {"status": "ok", "pool_id": p.id}


@admin_router.post("/pools/members")
def upsert_pool_member(
    body: PoolMemberUpsert,
    db: Session = Depends(get_session),
    current_account: dict = Depends(get_current_account),
):
    # Validate tenant access
    if not validate_tenant_access(body.tenant_id, current_account):
        raise HTTPException(403, "Access denied to this tenant")
    pid = f"{body.pool_id}:{body.user_id}"
    m = db.query(RoutingPoolMember).filter_by(id=pid).one_or_none()
    if not m:
        m = RoutingPoolMember(
            id=pid,
            pool_id=body.pool_id,
            tenant_id=body.tenant_id,
            user_id=body.user_id,
            weight=body.weight or 1,
        )
        db.add(m)
    else:
        m.weight = body.weight or 1
    db.commit()
    return {"status": "ok", "member": {"user_id": body.user_id, "weight": m.weight}}


@admin_router.post("/rules")
def upsert_rule(
    body: RuleUpsert,
    db: Session = Depends(get_session),
    current_account: dict = Depends(get_current_account),
):
    # Validate tenant access
    if not validate_tenant_access(body.tenant_id, current_account):
        raise HTTPException(403, "Access denied to this tenant")
    r = (
        db.query(RoutingRule)
        .filter_by(tenant_id=body.tenant_id, id=body.rule_id)
        .one_or_none()
    )
    if not r:
        r = RoutingRule(
            id=body.rule_id,
            tenant_id=body.tenant_id,
            name=body.name,
            priority=body.priority,
            active=body.active,
            condition=body.condition,
            action=body.action,
        )
        db.add(r)
    else:
        r.name, r.priority, r.active, r.condition, r.action = (
            body.name,
            body.priority,
            body.active,
            body.condition,
            body.action,
        )
    db.commit()
    return {"status": "ok", "rule_id": r.id}


# ----------------------- Enhanced Admin Endpoints -----------------------


@admin_router.post("/route/dry-run-batch")
async def dry_run_batch(
    tenant_id: str = Body(...),
    leads: Optional[List[LeadIn]] = Body(None),
    limit: int = Body(100),
    db: Session = Depends(get_session),
    current_account: dict = Depends(get_current_account),
):
    # Validate tenant access
    if not validate_tenant_access(tenant_id, current_account):
        raise HTTPException(403, "Access denied to this tenant")
    """Preview routing for multiple leads without committing decisions"""

    # If no leads provided, generate sample leads for demo
    if not leads:
        # In production, you would fetch unrouted leads from your lead table
        # For now, create sample leads for testing
        sample_companies = [
            ("john@acme.com", "Acme Corp", "acme.com"),
            ("jane@widget.io", "Widget Inc", "widget.io"),
            ("bob@startup.tech", "Startup Tech", "startup.tech"),
            ("alice@enterprise.com", "Enterprise Solutions", "enterprise.com"),
            ("mike@sales.biz", "Sales Pro", "sales.biz"),
        ]

        leads = []
        for i, (email, company, website) in enumerate(
            sample_companies[: min(limit, len(sample_companies))]
        ):
            leads.append(
                LeadIn(
                    lead_id=f"sample_{i+1}",
                    email=email,
                    company=company,
                    website=website,
                )
            )

    # Limit the number of leads to process
    leads = leads[:limit]

    # Get Salesforce adapter for routing
    try:
        sf = await get_salesforce_gateway_for_tenant(db, tenant_id=tenant_id)
        adapter = SalesforceAdapter(sf)
    except Exception as e:
        logger.warning("Failed to get Salesforce adapter for dry-run batch: %s", str(e))
        # Continue without adapter for demo purposes
        adapter = None

    # Process each lead through routing engine
    results = []
    warnings = []
    summary = {"assigned_owner": 0, "assigned_pool": 0, "no_match": 0}

    for lead in leads:
        try:
            # Route with preview=True to avoid persisting decisions
            decision = await route_lead(
                db, lead, tenant_id=tenant_id, crm_adapter=adapter, preview=True
            )

            # Add lead_id to result
            result = {"lead_id": lead.lead_id or "unknown", **decision}
            results.append(result)

            # Update summary counts
            if decision["decision"] == "assign_owner":
                summary["assigned_owner"] += 1
            elif decision["decision"] == "assign_pool":
                summary["assigned_pool"] += 1
            else:
                summary["no_match"] += 1

        except Exception as e:
            warning = {"lead_id": lead.lead_id or "unknown", "error": str(e)}
            warnings.append(warning)
            logger.warning("Error routing lead %s in dry-run: %s", lead.lead_id, str(e))

    return {
        "tenant_id": tenant_id,
        "total_processed": len(results),
        "results": results,
        "warnings": warnings,
        "summary": summary,
    }


@admin_router.get("/decisions")
async def list_decisions(
    tenant_id: str = Query(...),
    limit: int = Query(50),
    include_rolled_back: bool = Query(False),
    db: Session = Depends(get_session),
    current_account: dict = Depends(get_current_account),
):
    # Validate tenant access
    if not validate_tenant_access(tenant_id, current_account):
        raise HTTPException(403, "Access denied to this tenant")
    """Get recent routing decisions with filtering"""
    query = db.query(RoutingDecision).filter_by(tenant_id=tenant_id)

    if not include_rolled_back:
        query = query.filter(RoutingDecision.rolled_back_at.is_(None))

    decisions = query.order_by(RoutingDecision.created_at.desc()).limit(limit).all()

    return [
        {
            "id": d.id,
            "lead_id": d.lead_id,
            "decision": d.decision,
            "owner_id": d.owner_id,
            "pool_id": d.pool_id,
            "rule_id": d.rule_id,
            "reason": d.reason,
            "context": d.context,
            "created_at": d.created_at.isoformat() if d.created_at else None,
            "applied_at": d.applied_at.isoformat() if d.applied_at else None,
            "rolled_back_at": d.rolled_back_at.isoformat()
            if d.rolled_back_at
            else None,
        }
        for d in decisions
    ]


@admin_router.post("/decisions/{decision_id}/rollback")
async def rollback_decision(
    decision_id: str,
    tenant_id: str = Body(...),
    reason: str = Body(None),
    db: Session = Depends(get_session),
    current_account: dict = Depends(get_current_account),
):
    # Validate tenant access
    if not validate_tenant_access(tenant_id, current_account):
        raise HTTPException(403, "Access denied to this tenant")
    """Rollback a routing decision"""
    decision = (
        db.query(RoutingDecision).filter_by(id=decision_id, tenant_id=tenant_id).first()
    )

    if not decision:
        raise HTTPException(404, "Decision not found")

    if decision.rolled_back_at:
        raise HTTPException(400, "Decision already rolled back")

    # Mark as rolled back
    decision.rolled_back_at = datetime.utcnow()

    # If it was a pool assignment, decrement the counter with retry logic
    if decision.pool_id and decision.applied_at and decision.owner_id:
        max_retries = 3
        retry_count = 0

        while retry_count < max_retries:
            try:
                pool_state = (
                    db.query(RoutingPoolState)
                    .filter_by(id=decision.pool_id, tenant_id=tenant_id)
                    .with_for_update()
                    .first()
                )

                if pool_state:
                    assigned = pool_state.assigned_today or {}
                    if decision.owner_id in assigned:
                        assigned[decision.owner_id] = max(
                            0, assigned[decision.owner_id] - 1
                        )
                        pool_state.assigned_today = assigned
                        pool_state.updated_at = datetime.utcnow()
                        # version will be incremented automatically by SQLAlchemy

                db.commit()
                break

            except StaleDataError:
                # Concurrent modification detected, retry
                db.rollback()
                retry_count += 1
                if retry_count < max_retries:
                    time.sleep(0.1 * retry_count)  # Exponential backoff
                    logger.warning(
                        "Optimistic lock retry %d for rollback decision %s",
                        retry_count,
                        decision_id,
                    )
                else:
                    logger.error(
                        "Failed to rollback decision %s after %d retries",
                        decision_id,
                        max_retries,
                    )
                    raise HTTPException(
                        409, "Concurrent modification conflict, please retry"
                    )
    else:
        # No pool state to update, just commit the decision rollback
        db.commit()

    logger.info(
        "routing_rollback tenant=%s decision=%s reason=%s",
        tenant_id,
        decision_id,
        (reason or ""),
    )

    return {
        "status": "ok",
        "decision_id": decision_id,
        "rolled_back_at": decision.rolled_back_at.isoformat(),
    }


@admin_router.get("/pools/{pool_id}/health")
async def pool_health(
    pool_id: str,
    tenant_id: str = Query(...),
    db: Session = Depends(get_session),
    current_account: dict = Depends(get_current_account),
):
    # Validate tenant access
    if not validate_tenant_access(tenant_id, current_account):
        raise HTTPException(403, "Access denied to this tenant")
    """Get pool health metrics"""
    pool = db.query(RoutingPool).filter_by(id=pool_id, tenant_id=tenant_id).first()

    if not pool:
        raise HTTPException(404, "Pool not found")

    members = (
        db.query(RoutingPoolMember)
        .filter_by(pool_id=pool_id, tenant_id=tenant_id)
        .all()
    )

    state = (
        db.query(RoutingPoolState).filter_by(id=pool_id, tenant_id=tenant_id).first()
    )

    assigned_today = (state.assigned_today if state else {}) or {}

    # Calculate health metrics
    members_at_cap = 0
    total_assigned = 0
    member_stats = []

    for member in members:
        assigned = assigned_today.get(member.user_id, 0)
        total_assigned += assigned

        at_cap = False
        if pool.daily_cap and assigned >= pool.daily_cap:
            members_at_cap += 1
            at_cap = True

        member_stats.append(
            {
                "user_id": member.user_id,
                "weight": member.weight,
                "assigned_today": assigned,
                "at_capacity": at_cap,
                "utilization": (assigned / pool.daily_cap * 100)
                if pool.daily_cap
                else 0,
            }
        )

    # Calculate reset time
    reset_time = None
    if state and state.reset_at:
        # Calculate midnight in tenant's timezone
        tz = ZoneInfo(state.tenant_timezone or "UTC")
        now_tz = datetime.now(tz)
        tomorrow_local_midnight = (now_tz + timedelta(days=1)).replace(
            hour=0, minute=0, second=0, microsecond=0
        )
        # Convert to UTC for consistent storage/display
        reset_time = tomorrow_local_midnight.astimezone(ZoneInfo("UTC")).isoformat()

    # Calculate actual capacity utilization
    cap_util = 0
    if pool.daily_cap and members:
        cap_util = (total_assigned / (pool.daily_cap * len(members))) * 100

    return {
        "pool_id": pool_id,
        "name": pool.name,
        "algorithm": pool.algo,
        "daily_cap": pool.daily_cap,
        "member_count": len(members),
        "members_at_capacity": members_at_cap,
        "capacity_utilization": cap_util,
        "total_assigned_today": total_assigned,
        "average_assignments": total_assigned / len(members) if members else 0,
        "reset_time": reset_time,
        "member_stats": member_stats,
        "health_status": "healthy" if members_at_cap < len(members) else "at_capacity",
    }


@admin_router.post("/pools/{pool_id}/reset")
async def reset_pool_state(
    pool_id: str,
    tenant_id: str = Body(...),
    db: Session = Depends(get_session),
    current_account: dict = Depends(get_current_account),
):
    # Validate tenant access
    if not validate_tenant_access(tenant_id, current_account):
        raise HTTPException(403, "Access denied to this tenant")
    """Reset pool assignment counters"""
    max_retries = 3
    retry_count = 0

    while retry_count < max_retries:
        try:
            state = (
                db.query(RoutingPoolState)
                .filter_by(id=pool_id, tenant_id=tenant_id)
                .with_for_update()
                .first()
            )

            if not state:
                raise HTTPException(404, "Pool state not found")

            # Reset counters
            state.assigned_today = {}
            state.reset_at = datetime.utcnow()
            state.updated_at = datetime.utcnow()
            # version will be incremented automatically by SQLAlchemy

            db.commit()

            logger.info("pool_reset tenant=%s pool=%s", tenant_id, pool_id)

            return {
                "status": "ok",
                "pool_id": pool_id,
                "reset_at": state.reset_at.isoformat(),
            }

        except StaleDataError:
            # Concurrent modification detected, retry
            db.rollback()
            retry_count += 1
            if retry_count < max_retries:
                time.sleep(0.1 * retry_count)  # Exponential backoff
                logger.warning(
                    "Optimistic lock retry %d for reset pool %s", retry_count, pool_id
                )
            else:
                logger.error(
                    "Failed to reset pool %s after %d retries", pool_id, max_retries
                )
                raise HTTPException(
                    409, "Concurrent modification conflict, please retry"
                )



from __future__ import annotations

import json
import os
import re
from typing import Any, Dict, List

import pandas as pd

from ..services.engine_client import EngineClient
from ..api.intelligence_client import IntelligenceClient
from ..repos.tradeshow_repo import TradeshowRepo
from ..services.metrics import metrics
from ..model_defs.c2a_models import C2ASuggestion


class TradeshowWorker:
    def __init__(self, db, storage=None):
        self.db = db
        self.repo = TradeshowRepo(db)
        self.engine = EngineClient()
        self.intel = IntelligenceClient()
        self.storage = storage  # optional: if provided, must implement read(key)

    async def run_job(self, job_id: str) -> None:
        # Parse import_id from job_id pattern ts_{import_id}_{ts}
        m = re.match(r"^ts_(?P<imp>[a-f0-9]{16})_", job_id)
        if not m:
            return
        import_id = m.group("imp")
        row = await self.repo.get_import(import_id)
        if not row:
            return

        account_id = str(row["account_id"])
        artifact_key = row["artifact_key"]

        # Fetch policy thresholds from Intelligence
        policy = await self.intel.get_policy(account_id=account_id, segment=None)
        suggest = float(policy.get("suggest", 0.70))

        # Load dataset from storage
        df = await self._load_dataframe(artifact_key)

        # Build config using Intelligence auto-configure
        cfg = await self.intel.auto_configure(
            df, df, goal=os.getenv("TRADESHOW_GOAL", "match")
        )
        mappings = cfg.get("field_mappings") or cfg.get("mappings") or []

        # Chunking for scale
        try:
            chunk_rows = int(os.getenv("TRADESHOW_CHUNK_ROWS", "50000") or 50000)
        except Exception:
            chunk_rows = 50000

        total_rows = len(df)
        matched_count = 0
        last_result = None

        async def _emit_suggestions(matches_df):
            suggestions: List[C2ASuggestion] = []
            for idx, r in matches_df.iterrows():
                src_id = r.get("source_id") or r.get("src_id") or str(idx)
                fsd = (
                    r.get("field_score_details")
                    or r.get("feature_scores")
                    or r.get("field_scores")
                    or {}
                )
                reasons = self._top_reasons(fsd)
                s = C2ASuggestion(
                    policy_id=row.get("config_version") or "v0",
                    contact_id=str(src_id),
                    account_id=str(account_id),
                    job_id=job_id,
                    score=float(r.get("score") or r.get("match_score") or 0.0),
                    reasons=reasons,
                    field_score_details=fsd if isinstance(fsd, (dict, list)) else None,
                    explanation="tradeshow",
                )
                suggestions.append(s)
                if len(suggestions) >= 500:
                    self.db.add_all(suggestions)
                    await self.db.commit()
                    suggestions = []
            if suggestions:
                self.db.add_all(suggestions)
                await self.db.commit()

        if total_rows <= chunk_rows:
            last_result = await self.engine.match_records(
                df,
                df,
                mappings=mappings,
                threshold=suggest,
                job_id=job_id,
            )
            matched_count = int(len(last_result.matches))
            await _emit_suggestions(last_result.matches)
        else:
            processed = 0
            for start in range(0, total_rows, chunk_rows):
                end = min(start + chunk_rows, total_rows)
                chunk_df = df.iloc[start:end]
                res = await self.engine.match_records(
                    chunk_df,
                    df,
                    mappings=mappings,
                    threshold=suggest,
                    job_id=job_id,
                )
                last_result = res
                matched_count += int(len(res.matches))
                await _emit_suggestions(res.matches)
                processed = end
                pct = int(processed * 100 / max(1, total_rows))
                try:
                    await self.repo.save_report(
                        job_id=job_id,
                        account_id=account_id,
                        matched_count=matched_count,
                        new_count=max(0, processed - matched_count),
                        undo_token="n/a",
                        summary={"status": "running", "pct_complete": pct},
                    )
                except Exception:
                    pass

        new_count = max(0, total_rows - matched_count)

        # Final report save
        summary = {
            "engine_version": getattr(last_result, "engine_version", ""),
            "processing_ms": getattr(last_result, "processing_time_ms", 0),
            "config_hash": getattr(last_result, "config_hash", ""),
            "field_score_details": "histogram_omitted",
            "status": "completed",
            "pct_complete": 100,
        }
        await self.repo.save_report(
            job_id=job_id,
            account_id=account_id,
            matched_count=matched_count,
            new_count=new_count,
            undo_token="n/a",
            summary=summary,
        )

        try:
            await metrics.increment("job_completed")
            await metrics.increment("suggestions_emitted", by=matched_count + new_count)
        except Exception:
            pass

    async def _load_dataframe(self, key: str) -> pd.DataFrame:
        # Local storage: LocalStorageBackend used by TradeshowService has read()
        if self.storage and hasattr(self.storage, "read"):
            content = await self.storage.read(key)
            from io import BytesIO

            # Detect file type
            lower = (key or "").lower()
            if lower.endswith((".xlsx", ".xls")):
                return pd.read_excel(BytesIO(content))
            return pd.read_csv(BytesIO(content))
        # Fallback: try to use core io loader if available
        try:
            from fmatch.core.io import FileLoader

            loader = FileLoader()
            lazy = loader.scan_csv(key)
            return lazy.collect().to_pandas()
        except Exception:
            # Last resort: attempt local path
            import os

            if os.path.exists(key):
                if key.lower().endswith((".xlsx", ".xls")):
                    return pd.read_excel(key)
                return pd.read_csv(key)
            raise

    def _top_reasons(self, field_score_details: Any) -> Dict[str, Any]:
        # Summarize top contributions for explainability
        try:
            if isinstance(field_score_details, str):
                data = json.loads(field_score_details)
            else:
                data = field_score_details
            if isinstance(data, list) and data:
                data = data[:3]
            return {"top": data}
        except Exception:
            return {}



from __future__ import annotations

import asyncio
import hashlib
import json
import os
import time
from dataclasses import dataclass
import math
import re
from typing import Any, Dict, Optional, Tuple

import redis.asyncio as aioredis
from fastapi import Depends, Header, HTTPException, Request
from fastapi.security import HTTPAuthorizationCredentials
from sqlalchemy.ext.asyncio import AsyncSession

from ...db import get_session
from ...settings import (
    API_QPS_DEFAULT,
    REDIS_URL,
    DEV_ACCOUNT_ID,
)
from ...repos.api_repo import ApiRepo, ApiContext
from ...model_defs.api_models import ApiUsage
from sqlalchemy import select


# Simple in-proc token bucket per key and per IP
class _TokenBucket:
    def __init__(self, rate_per_sec: float, capacity: int):
        self.rate = float(rate_per_sec)
        self.capacity = int(max(1, capacity))
        self.tokens = float(self.capacity)
        self.last = time.monotonic()

    def allow(self) -> bool:
        now = time.monotonic()
        elapsed = now - self.last
        self.last = now
        self.tokens = min(self.capacity, self.tokens + elapsed * self.rate)
        if self.tokens >= 1.0:
            self.tokens -= 1.0
            return True
        return False

    def retry_after_seconds(self) -> int:
        """Approximate seconds until 1 token is available."""
        if self.tokens >= 1.0:
            return 0
        need = 1.0 - max(0.0, self.tokens)
        if self.rate <= 0:
            return 1
        return max(1, math.ceil(need / self.rate))

    def remaining_int(self) -> int:
        return max(0, int(self.tokens))


_buckets_key: Dict[str, _TokenBucket] = {}
_buckets_ip: Dict[str, _TokenBucket] = {}


def _qps_for_tier(tier: str) -> int:
    tier = (tier or "free").upper()
    return int(API_QPS_DEFAULT.get(tier, API_QPS_DEFAULT.get("FREE", 2)))


async def _get_redis() -> Optional[aioredis.Redis]:
    url = os.getenv("REDIS_URL", REDIS_URL or "")
    if not url:
        return None
    try:
        return aioredis.from_url(url, decode_responses=True)
    except Exception:
        return None


# Idempotency store (local + optional Redis)
class IdempotencyStore:
    def __init__(self):
        self._local: Dict[str, Tuple[float, Dict[str, Any]]] = {}
        self._ttl = int(os.getenv("API_IDEMPOTENCY_TTL_SEC", "600") or 600)
        self._lock = asyncio.Lock()

    def _key(self, api_key_prefix: str, idem_key: str) -> str:
        return f"idem:{api_key_prefix}:{hashlib.md5((idem_key or '').encode()).hexdigest()}"

    async def get(
        self, api_key_prefix: str, idem_key: Optional[str]
    ) -> Optional[Dict[str, Any]]:
        if not idem_key:
            return None
        # Try Redis first
        r = await _get_redis()
        cache_key = self._key(api_key_prefix, idem_key)
        if r:
            try:
                raw = await r.get(cache_key)
                if raw:
                    return json.loads(raw)
            except Exception:
                pass
        # Fallback local
        async with self._lock:
            item = self._local.get(cache_key)
            if not item:
                return None
            ts, payload = item
            if time.time() - ts > self._ttl:
                self._local.pop(cache_key, None)
                return None
            return payload

    async def set(
        self, api_key_prefix: str, idem_key: Optional[str], payload: Dict[str, Any]
    ) -> None:
        if not idem_key:
            return
        cache_key = self._key(api_key_prefix, idem_key)
        ttl = self._ttl
        # Redis
        r = await _get_redis()
        if r:
            try:
                await r.set(cache_key, json.dumps(payload), ex=ttl)
                return
            except Exception:
                pass
        # Local
        async with self._lock:
            self._local[cache_key] = (time.time(), payload)


idem_store = IdempotencyStore()


@dataclass
class ApiRequestContext:
    api_key_prefix: str
    account_id: str
    tier: str
    qps_limit: int


_AUTH_RX = re.compile(r"^\s*(Bearer|Api-Key)\s+(.+)$", re.I)


async def require_api_key(
    request: Request,
    authorization: Optional[str] = Header(default=None, alias="Authorization"),
    x_api_key: Optional[str] = Header(default=None, alias="X-API-Key"),
    db: AsyncSession = Depends(get_session),
) -> ApiRequestContext:
    # Dev bypass: allow requests without DB/key only in explicit dev environments,
    # and only when explicitly enabled (opt-in).
    env = (os.getenv("ENV") or "").strip().lower()
    if env in ("dev", "development") and os.getenv("FM_BYPASS_API_KEY", "0") == "1":
        # Populate request.state with reasonable defaults for downstream
        request.state.account_id = DEV_ACCOUNT_ID
        request.state.api_key_prefix = "dev"
        request.state.tier = "DEV"
        try:
            request.state.x_ratelimit_limit = 1000
            request.state.x_ratelimit_remaining = 999
        except Exception:
            pass
        return ApiRequestContext(
            api_key_prefix="dev",
            account_id=DEV_ACCOUNT_ID,
            tier="DEV",
            qps_limit=1000,
        )

    # Accept either X-API-Key or Authorization (Bearer|Api-Key <token>)
    token: str = ""
    scheme: Optional[str] = None
    if x_api_key:
        token = (x_api_key or "").strip()
        scheme = "api-key"
    elif authorization:
        m = _AUTH_RX.match(authorization or "")
        scheme = (m.group(1).lower() if m else None) if authorization else None
        token = (m.group(2) if m else authorization).strip()
    if not token:
        raise HTTPException(
            status_code=401,
            detail={"error": "UNAUTHORIZED", "message": "Missing API key"},
        )

    # Bearer JWT support (Sheets/Clerk tokens) for v1 endpoints
    if scheme == "bearer" and not token.startswith("fm_"):
        from ...auth_core import get_current_identity

        ident = await get_current_identity(
            HTTPAuthorizationCredentials(scheme="Bearer", credentials=token)
        )
        account_id = ident.get("org_id")
        if not account_id:
            raise HTTPException(
                status_code=401,
                detail={"error": "UNAUTHORIZED", "message": "No organization on token"},
            )

        api_key_prefix = f"jwt:{account_id}"
        tier = "JWT"
        qps_limit = _qps_for_tier(tier)

        request.state.account_id = account_id
        request.state.api_key_prefix = api_key_prefix
        request.state.tier = tier
        try:
            request.state.x_ratelimit_limit = qps_limit
            request.state.x_ratelimit_remaining = qps_limit
        except Exception:
            pass

        return ApiRequestContext(
            api_key_prefix=api_key_prefix,
            account_id=account_id,
            tier=tier,
            qps_limit=qps_limit,
        )

    repo = ApiRepo(db)
    ctx: Optional[ApiContext] = await repo.get_key(token)
    if not ctx:
        raise HTTPException(
            status_code=401,
            detail={"error": "UNAUTHORIZED", "message": "Invalid API key"},
        )
    if ctx.status != "active":
        raise HTTPException(
            status_code=401,
            detail={"error": "UNAUTHORIZED", "message": "API key disabled"},
        )

    # Quota enforcement (monthly)
    if ctx.monthly_quota is not None and ctx.monthly_quota >= 0:
        now = time.gmtime()
        period = now.tm_year * 100 + now.tm_mon
        stmt = select(ApiUsage).where(
            ApiUsage.api_key == ctx.api_key_prefix, ApiUsage.period_yyyymm == period
        )
        r = await db.execute(stmt)
        usage = r.scalar_one_or_none()
        if usage and usage.calls is not None and usage.calls >= int(ctx.monthly_quota):
            raise HTTPException(
                status_code=429,
                detail={"error": "RATE_LIMIT", "message": "Monthly quota exceeded"},
            )

    # Per-key rate limiting (token bucket)
    key = f"key:{ctx.api_key_prefix}"
    bucket = _buckets_key.get(key)
    key_qps = int(ctx.qps_limit or _qps_for_tier(ctx.tier))
    if not bucket:
        bucket = _TokenBucket(rate_per_sec=key_qps, capacity=max(1, key_qps))
        _buckets_key[key] = bucket
    if not bucket.allow():
        headers = {
            "Retry-After": str(bucket.retry_after_seconds()),
            "X-RateLimit-Limit": str(key_qps),
            "X-RateLimit-Remaining": "0",
        }
        raise HTTPException(
            status_code=429,
            detail={"error": "RATE_LIMIT", "message": "Per-key rate limit exceeded"},
            headers=headers,
        )

    # Per-IP rate limiting with tier defaults
    ip = request.client.host if request.client else "unknown"
    ipq = _qps_for_tier(ctx.tier)
    ip_bucket = _buckets_ip.get(ip)
    if not ip_bucket:
        ip_bucket = _TokenBucket(rate_per_sec=ipq, capacity=max(1, ipq))
        _buckets_ip[ip] = ip_bucket
    if not ip_bucket.allow():
        headers = {
            "Retry-After": str(ip_bucket.retry_after_seconds()),
            "X-RateLimit-Limit": str(ipq),
            "X-RateLimit-Remaining": "0",
        }
        raise HTTPException(
            status_code=429,
            detail={"error": "RATE_LIMIT", "message": "Per-IP rate limit exceeded"},
            headers=headers,
        )

    # Attach to request.state for downstream
    request.state.account_id = ctx.account_id
    request.state.api_key_prefix = ctx.api_key_prefix
    request.state.tier = ctx.tier

    # Expose rate-limit hints for middleware headers (per key)
    try:
        request.state.x_ratelimit_limit = key_qps
        request.state.x_ratelimit_remaining = bucket.remaining_int()
    except Exception:
        pass

    return ApiRequestContext(
        api_key_prefix=ctx.api_key_prefix,
        account_id=ctx.account_id,
        tier=ctx.tier,
        qps_limit=key_qps,
    )



"""SQLite-backed state store for Architect projects."""

from __future__ import annotations

import json
import sqlite3
import uuid
from contextlib import contextmanager
from datetime import datetime, timezone
from pathlib import Path
from typing import Iterator, List

from g_gremlin.paths import gremlin_path

from .models import ArchitectProject, ProjectPhaseResult, ProjectStatus


def _utc_now() -> datetime:
    return datetime.now(timezone.utc)


def _utc_now_iso() -> str:
    return _utc_now().isoformat()


class ArchitectStateStore:
    def __init__(self, base_dir: Path | None = None) -> None:
        self.base_dir = Path(base_dir) if base_dir else gremlin_path("architect")
        self.projects_dir = self.base_dir / "projects"
        self.db_path = self.base_dir / "db" / "architect.sqlite"
        self.projects_dir.mkdir(parents=True, exist_ok=True)
        self.db_path.parent.mkdir(parents=True, exist_ok=True)
        self._init_db()

    @contextmanager
    def _connect(self) -> Iterator[sqlite3.Connection]:
        conn = sqlite3.connect(self.db_path)
        conn.row_factory = sqlite3.Row
        try:
            yield conn
            conn.commit()
        except Exception:
            conn.rollback()
            raise
        finally:
            conn.close()

    def _init_db(self) -> None:
        with self._connect() as conn:
            conn.execute(
                """
                CREATE TABLE IF NOT EXISTS architect_projects (
                    project_id TEXT PRIMARY KEY,
                    blueprint_name TEXT NOT NULL,
                    status TEXT NOT NULL,
                    parameters_json TEXT NOT NULL,
                    project_dir TEXT NOT NULL,
                    phase_results_json TEXT,
                    proposal_hash TEXT,
                    sandbox_org_alias TEXT NOT NULL,
                    prod_org_alias TEXT NOT NULL,
                    variant TEXT,
                    prod_only INTEGER DEFAULT 0,
                    receipt_ids_json TEXT,
                    created_at TEXT NOT NULL,
                    updated_at TEXT NOT NULL
                )
                """
            )
            self._ensure_column(conn, "variant", "TEXT")
            self._ensure_column(conn, "prod_only", "INTEGER DEFAULT 0")

    def _ensure_column(self, conn: sqlite3.Connection, name: str, ddl: str) -> None:
        try:
            conn.execute(f"ALTER TABLE architect_projects ADD COLUMN {name} {ddl}")
        except sqlite3.OperationalError:
            return

    def create_project(
        self,
        *,
        blueprint_name: str,
        parameters: dict,
        sandbox_org_alias: str,
        prod_org_alias: str,
        variant: str | None = None,
        prod_only: bool = False,
    ) -> ArchitectProject:
        project_id = f"arch_{uuid.uuid4().hex[:12]}"
        project_dir = self.projects_dir / project_id
        self._init_project_dirs(project_dir)

        project = ArchitectProject(
            project_id=project_id,
            blueprint_name=blueprint_name,
            status=ProjectStatus.CREATED,
            parameters=parameters,
            project_dir=str(project_dir),
            sandbox_org_alias=sandbox_org_alias,
            prod_org_alias=prod_org_alias,
            phase_results={},
            receipt_ids={},
            variant=variant,
            prod_only=prod_only,
        )

        now = _utc_now_iso()
        with self._connect() as conn:
            conn.execute(
                """
                INSERT INTO architect_projects (
                    project_id, blueprint_name, status, parameters_json, project_dir,
                    phase_results_json, proposal_hash, sandbox_org_alias, prod_org_alias,
                    variant, prod_only, receipt_ids_json, created_at, updated_at
                )
                VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)
                """,
                (
                    project.project_id,
                    project.blueprint_name,
                    project.status.value,
                    json.dumps(project.parameters, ensure_ascii=True),
                    project.project_dir,
                    json.dumps({}, ensure_ascii=True),
                    project.proposal_hash,
                    project.sandbox_org_alias,
                    project.prod_org_alias,
                    project.variant,
                    1 if project.prod_only else 0,
                    json.dumps(project.receipt_ids, ensure_ascii=True),
                    now,
                    now,
                ),
            )

        self._write_project_json(project)
        return project

    def get_project(self, project_id: str) -> ArchitectProject | None:
        with self._connect() as conn:
            row = conn.execute(
                "SELECT * FROM architect_projects WHERE project_id = ?",
                (project_id,),
            ).fetchone()
        if not row:
            return None
        return self._row_to_project(row)

    def list_projects(self) -> List[ArchitectProject]:
        with self._connect() as conn:
            rows = conn.execute(
                "SELECT * FROM architect_projects ORDER BY created_at DESC",
            ).fetchall()
        return [self._row_to_project(row) for row in rows]

    def update_status(self, project_id: str, status: ProjectStatus) -> ArchitectProject:
        project = self.get_project(project_id)
        if not project:
            raise ValueError(f"Project not found: {project_id}")
        project.status = status
        self._persist_project(project)
        return project

    def record_phase_result(self, project_id: str, result: ProjectPhaseResult) -> ArchitectProject:
        project = self.get_project(project_id)
        if not project:
            raise ValueError(f"Project not found: {project_id}")
        project.phase_results[result.phase] = result
        self._persist_project(project)
        return project

    def update_project(self, project: ArchitectProject) -> ArchitectProject:
        self._persist_project(project)
        return project

    def _persist_project(self, project: ArchitectProject) -> None:
        now = _utc_now_iso()
        with self._connect() as conn:
            conn.execute(
                """
                UPDATE architect_projects
                SET status = ?,
                    parameters_json = ?,
                    project_dir = ?,
                    phase_results_json = ?,
                    proposal_hash = ?,
                    sandbox_org_alias = ?,
                    prod_org_alias = ?,
                    variant = ?,
                    prod_only = ?,
                    receipt_ids_json = ?,
                    updated_at = ?
                WHERE project_id = ?
                """,
                (
                    project.status.value,
                    json.dumps(project.parameters, ensure_ascii=True),
                    project.project_dir,
                    json.dumps(
                        {k: v.model_dump() for k, v in project.phase_results.items()},
                        ensure_ascii=True,
                    ),
                    project.proposal_hash,
                    project.sandbox_org_alias,
                    project.prod_org_alias,
                    project.variant,
                    1 if project.prod_only else 0,
                    json.dumps(project.receipt_ids, ensure_ascii=True),
                    now,
                    project.project_id,
                ),
            )
        self._write_project_json(project)

    def _row_to_project(self, row: sqlite3.Row) -> ArchitectProject:
        phase_results_raw = json.loads(row["phase_results_json"] or "{}")
        phase_results = {
            key: ProjectPhaseResult.model_validate(value)
            for key, value in phase_results_raw.items()
        }
        receipt_ids = json.loads(row["receipt_ids_json"] or "{}")
        status_value = row["status"]
        if status_value == "promoting_prod":
            status_value = ProjectStatus.DEPLOYING_PROD.value
        elif status_value == "promoted_prod":
            status_value = ProjectStatus.DEPLOYED_PROD.value
        return ArchitectProject(
            project_id=row["project_id"],
            blueprint_name=row["blueprint_name"],
            status=ProjectStatus(status_value),
            parameters=json.loads(row["parameters_json"] or "{}"),
            project_dir=row["project_dir"],
            phase_results=phase_results,
            proposal_hash=row["proposal_hash"],
            sandbox_org_alias=row["sandbox_org_alias"],
            prod_org_alias=row["prod_org_alias"],
            receipt_ids=receipt_ids,
            variant=row["variant"],
            prod_only=bool(row["prod_only"] or 0),
        )

    def _init_project_dirs(self, project_dir: Path) -> None:
        project_dir.mkdir(parents=True, exist_ok=True)
        (project_dir / "discovery" / "raw").mkdir(parents=True, exist_ok=True)
        (project_dir / "proposal" / "metadata_pack").mkdir(parents=True, exist_ok=True)
        (project_dir / "proposal" / "hubspot_pack").mkdir(parents=True, exist_ok=True)
        (project_dir / "verify_sandbox").mkdir(parents=True, exist_ok=True)
        (project_dir / "deploy_prod").mkdir(parents=True, exist_ok=True)
        (project_dir / "promote_prod").mkdir(parents=True, exist_ok=True)

    def _write_project_json(self, project: ArchitectProject) -> None:
        path = Path(project.project_dir) / "project.json"
        payload = project.model_dump()
        path.write_text(json.dumps(payload, indent=2, ensure_ascii=True), encoding="utf-8")


__all__ = ["ArchitectStateStore"]



from __future__ import annotations

from fastapi import APIRouter, Depends
from pydantic import BaseModel
from typing import Optional, Dict, Any

from ...auth_security import require_admin, require_account
from ...db import get_session
from sqlalchemy.ext.asyncio import AsyncSession

from ...intelligence.decision_logger import DecisionLogger
from ...intelligence.bandit_repo import BanditRepo
from ...intelligence.policy_store import PolicyStore
from ...services.learning_pipeline import LearningPipeline


router = APIRouter(prefix="/api/v2/intelligence", tags=["Intelligence"])


class ResyncBody(BaseModel):
    account_id: Optional[str] = None
    segment: Optional[str] = None
    lookback_days: int = 30


@router.post("/learn/resync", dependencies=[Depends(require_admin)])
async def learn_resync(
    body: ResyncBody,
    db: AsyncSession = Depends(get_session),
    account: str = Depends(require_account),
) -> Dict[str, Any]:
    account_id = body.account_id or account
    # Assemble pipeline components
    decision_logger = DecisionLogger()
    bandits = BanditRepo()
    policies = PolicyStore(db)
    pipeline = LearningPipeline(decision_logger, bandits, policies)

    result = await pipeline.update_from_feedback(
        account_id=str(account_id),
        segment=body.segment,
        lookback_days=body.lookback_days,
    )
    return result



from __future__ import annotations

from pathlib import Path
from typing import Any, Dict

import yaml


def load_profile_by_name(profile_name: str) -> Dict[str, Any]:
    """Load a scoring profile by name from configs/scorer/profiles."""
    root = Path(__file__).parents[3] / "configs" / "scorer" / "profiles"
    path = root / f"{profile_name}.yaml"
    if not path.exists():
        raise FileNotFoundError(f"Scoring profile not found: {profile_name}")
    with path.open("r", encoding="utf-8") as handle:
        return yaml.safe_load(handle)



from __future__ import annotations

from typing import Any, Dict, List, Tuple, Optional
import os
import time
import httpx


class _TokenBucket:
    """Simple token bucket for rate limiting (best-effort, thread-safe-enough for our use).
    capacity tokens, refill per second.
    """

    def __init__(self, capacity: int, refill_rate_per_sec: float) -> None:
        self.capacity = max(1, capacity)
        self.tokens = float(self.capacity)
        self.refill = float(refill_rate_per_sec)
        self._last = time.time()

    def wait_for_token(self) -> None:
        now = time.time()
        elapsed = max(0.0, now - self._last)
        self._last = now
        self.tokens = min(self.capacity, self.tokens + elapsed * self.refill)
        if self.tokens < 1.0:
            # sleep until 1 token is available
            need = 1.0 - self.tokens
            delay = need / max(0.001, self.refill)
            time.sleep(delay)
            self.tokens = 0.0
            self._last = time.time()
        else:
            self.tokens -= 1.0


class ApolloAdapter:
    """Apollo enrichment adapter (stubbed calls, real retries/backoff scaffolding).

    Configure with env:
      APOLLO_API_KEY
      APOLLO_BASE_URL (optional; default https://api.apollo.io/v1)
      APOLLO_RPS (optional; default 3)
    """

    def __init__(
        self,
        api_key: Optional[str] = None,
        base_url: Optional[str] = None,
        rps: Optional[int] = None,
    ) -> None:
        self.api_key = api_key or os.getenv("APOLLO_API_KEY", "")
        self.base_url = (
            base_url or os.getenv("APOLLO_BASE_URL", "https://api.apollo.io/v1")
        ).rstrip("/")
        self.rps = int(os.getenv("APOLLO_RPS", str(rps or 3)))
        self._bucket = _TokenBucket(
            capacity=max(1, self.rps * 2), refill_rate_per_sec=float(self.rps)
        )
        self._client = httpx.Client(timeout=20.0)

    def _headers(self) -> Dict[str, str]:
        return {"Authorization": f"Bearer {self.api_key}", "Accept": "application/json"}

    def available(self) -> bool:
        return bool(self.api_key)

    def enrich_batch(
        self,
        headers: List[str],
        rows: List[List[Any]],
        fields: List[str],
    ) -> Tuple[List[List[Any]], float]:
        """Call Apollo in batches. Returns (enriched_rows, credits_used).
        Expected Apollo response shape:
          { "headers": [...], "rows": [[...], ...], "credits_used": 12.3 }
        """
        if not self.available():
            return rows, 0.0

        payload: Dict[str, Any] = {"headers": headers, "rows": rows, "fields": fields}

        self._bucket.wait_for_token()
        try:
            resp = self._client.post(
                f"{self.base_url}/enrich/batch",
                json=payload,
                headers=self._headers(),
            )
            resp.raise_for_status()
            data = resp.json() or {}
        except httpx.HTTPError:
            # Best-effort fallback: pass-through
            return rows, 0.0

        ap_headers = data.get("headers") or headers
        ap_rows = data.get("rows") or rows
        credits = float(data.get("credits_used") or 0.0)

        mapped_rows, _ = _map_apollo_rows(headers, ap_headers, ap_rows)
        return mapped_rows or rows, credits


def _map_apollo_rows(
    orig_headers: List[str], ap_headers: List[str], ap_rows: List[List[Any]]
) -> Tuple[List[List[Any]], List[str]]:
    """Align Apollo rows to original column order and append any new fields at the end."""
    ap_idx: Dict[str, int] = {str(h or "").strip(): i for i, h in enumerate(ap_headers)}
    out_headers: List[str] = list(orig_headers)
    # Append new fields once
    for h in ap_headers:
        if h not in out_headers:
            out_headers.append(h)

    out_rows: List[List[Any]] = []
    for r in ap_rows:
        base = [""] * len(out_headers)
        # fill original columns
        for i, h in enumerate(orig_headers):
            j = ap_idx.get(h)
            base[i] = r[j] if (j is not None and j < len(r)) else ""
        # fill appended columns
        for h in ap_headers:
            if h not in orig_headers:
                j = ap_idx.get(h)
                base[out_headers.index(h)] = (
                    r[j] if (j is not None and j < len(r)) else ""
                )
        out_rows.append(base)
    return out_rows, out_headers



"""
Documentation sink providers (Notion, Outline, Confluence).

Provides lightweight DocsSinkProvider implementations for idempotent page upserts
with conservative rate limit handling.
"""

from __future__ import annotations

import html as html_lib
import logging
import re
import time
from dataclasses import asdict, dataclass, field
from enum import Enum
from typing import Any, Dict, Iterable, List, Mapping, Optional, Protocol, Tuple

import requests

from g_gremlin.contracts import BatchOutcome


NOTION_API_BASE = "https://api.notion.com/v1"
NOTION_VERSION = "2022-06-28"
NOTION_LARGE_REPLACE_WARNING_BLOCKS = 150
OUTLINE_DEFAULT_API_BASE = "https://app.getoutline.com/api"
OUTLINE_EXTERNAL_ID_MARKER = "external_id:"
OUTLINE_METADATA_SENTINEL = "<!-- g-gremlin:metadata -->"
CONFLUENCE_DEFAULT_API_SUFFIX = "/wiki/rest/api"
CONFLUENCE_EXTERNAL_ID_KEY = "g_gremlin_external_id"
CONFLUENCE_MARKER_PREFIX = "g-gremlin external_id: "
CONFLUENCE_VOID_TAGS = (
    "area",
    "base",
    "br",
    "col",
    "hr",
    "img",
    "input",
    "link",
    "meta",
    "param",
    "source",
    "track",
    "wbr",
)

LOGGER = logging.getLogger(__name__)


class DocsProvider(str, Enum):
    NOTION = "notion"
    OUTLINE = "outline"
    WORD = "word"
    CONFLUENCE = "confluence"


@dataclass
class DocsPage:
    """Page to create or update."""

    external_id: str
    title: str
    content_markdown: str
    parent_id: Optional[str]
    properties: Dict[str, Any] = field(default_factory=dict)
    tags: List[str] = field(default_factory=list)


@dataclass
class DocsPublishResult:
    """Per-page publish outcome."""

    external_id: str
    status: str
    page_id: Optional[str]
    page_url: Optional[str]
    error: Optional[str] = None

    def to_dict(self) -> Dict[str, Any]:
        return asdict(self)


@dataclass
class DocsPublishLogEntry:
    external_id: str
    correlation_id: str
    provider: str
    page_id: Optional[str]
    page_url: Optional[str]
    status: str
    error_message: Optional[str] = None
    created_at: float = field(default_factory=lambda: time.time())


class DocsSinkProvider(Protocol):
    """Interface for documentation sinks."""

    @property
    def name(self) -> str:
        ...

    @property
    def provider_type(self) -> DocsProvider:
        ...

    def test_connection(self) -> Dict[str, Any]:
        ...

    def upsert_page(self, page: DocsPage) -> DocsPublishResult:
        ...

    def upsert_pages_batch(self, pages: List[DocsPage]) -> BatchOutcome:
        ...

    def get_page_by_external_id(self, external_id: str) -> Optional[Dict[str, Any]]:
        ...

    def delete_page(self, page_id: str) -> bool:
        ...


class DocsSinkError(RuntimeError):
    """Generic docs sink error."""


def _xhtml_clean(html: str) -> str:
    if not html:
        return html
    tag_group = "|".join(CONFLUENCE_VOID_TAGS)
    pattern = re.compile(rf"<({tag_group})(\\s[^>]*?)?>", re.IGNORECASE)

    def _self_close(match: re.Match[str]) -> str:
        raw = match.group(0)
        if raw.endswith("/>"):
            return raw
        return raw[:-1] + " />"

    return pattern.sub(_self_close, html)


def _slugify_label(tag: str) -> str:
    cleaned = re.sub(r"[^a-z0-9_-]+", "-", tag.strip().lower())
    cleaned = re.sub(r"-+", "-", cleaned).strip("-")
    return cleaned[:255]


def _build_markdown_renderer() -> "MarkdownIt":
    try:
        from markdown_it import MarkdownIt
    except ImportError as exc:  # pragma: no cover - optional dependency
        raise DocsSinkError(
            "markdown-it-py is required for Confluence markdown rendering. "
            "Install with: pip install g-gremlin[docs-confluence]"
        ) from exc

    md = MarkdownIt("commonmark", {"html": True, "linkify": True, "xhtmlOut": True})

    try:
        from mdit_py_plugins.container import container_plugin
    except ImportError as exc:  # pragma: no cover - optional dependency
        raise DocsSinkError(
            "mdit-py-plugins is required for Confluence containers. "
            "Install with: pip install g-gremlin[docs-confluence]"
        ) from exc

    def _container_render(self, tokens, idx, options, env):  # type: ignore[no-untyped-def]
        token = tokens[idx]
        info = token.info.strip() if token.info else ""
        if token.nesting == 1:
            if info.startswith("expand"):
                title = info[len("expand"):].strip()
                title_attr = f' data-title="{html_lib.escape(title)}"' if title else ""
                return f'<div class="expand"{title_attr}>'
            cls = info.split()[0].strip() if info else "info"
            return f'<div class="{cls}">'
        return "</div>"

    for name in ("info", "note", "tip", "warning", "caution", "important"):
        md.use(container_plugin, name, render=_container_render)

    md.use(
        container_plugin,
        "expand",
        render=_container_render,
        validate=lambda params, markup: params.strip().startswith("expand"),
    )

    return md


def _transform_container_divs(html: str) -> str:
    mapping = {
        "info": "info",
        "note": "note",
        "tip": "tip",
        "warning": "warning",
        "caution": "warning",
        "important": "note",
    }

    for cls, macro in mapping.items():
        pattern = re.compile(rf'<div class="{cls}">\s*(.*?)\s*</div>', re.DOTALL)

        def _wrap(match: re.Match[str]) -> str:
            inner = match.group(1)
            return (
                f'<ac:structured-macro ac:name="{macro}">'
                f'<ac:rich-text-body>{inner}</ac:rich-text-body>'
                "</ac:structured-macro>"
            )

        html = pattern.sub(_wrap, html)

    expand_pattern = re.compile(r'<div class="expand"(?:\s+data-title="([^"]*)")?>\s*(.*?)\s*</div>', re.DOTALL)

    def _expand_wrap(match: re.Match[str]) -> str:
        title = match.group(1)
        inner = match.group(2)
        title_param = f'<ac:parameter ac:name="title">{html_lib.escape(title)}</ac:parameter>' if title else ""
        return (
            '<ac:structured-macro ac:name="expand">'
            f"{title_param}"
            f'<ac:rich-text-body>{inner}</ac:rich-text-body>'
            "</ac:structured-macro>"
        )

    html = expand_pattern.sub(_expand_wrap, html)
    return html


def _transform_blockquotes(html: str) -> str:
    def _replace(match: re.Match[str]) -> str:
        inner = match.group(1).strip()
        first_para_match = re.match(r"<p>(.*?)</p>(.*)", inner, re.DOTALL)
        macro = "info"
        if first_para_match:
            first_para = first_para_match.group(1).strip()
            rest = first_para_match.group(2).strip()
            strong_match = re.match(r"^\s*<strong>([^<:]+):</strong>\s*(.*)$", first_para, re.DOTALL)
            if strong_match:
                label = strong_match.group(1).strip().lower()
                remainder = strong_match.group(2).strip()
                macro_map = {
                    "note": "note",
                    "important": "note",
                    "warning": "warning",
                    "caution": "warning",
                    "tip": "tip",
                }
                macro = macro_map.get(label, "info")
                rebuilt = []
                if remainder:
                    rebuilt.append(f"<p>{remainder}</p>")
                if rest:
                    rebuilt.append(rest)
                inner = "\n".join(rebuilt) if rebuilt else ""
        return (
            f'<ac:structured-macro ac:name="{macro}">'
            f'<ac:rich-text-body>{inner}</ac:rich-text-body>'
            "</ac:structured-macro>"
        )

    return re.sub(r"<blockquote>\s*(.*?)\s*</blockquote>", _replace, html, flags=re.DOTALL)


def _transform_status_lozenges(html: str) -> str:
    placeholders: List[str] = []

    def _stash(match: re.Match[str]) -> str:
        placeholders.append(match.group(0))
        return f"__GG_CODE_BLOCK_{len(placeholders) - 1}__"

    html = re.sub(r"<pre[^>]*>.*?</pre>", _stash, html, flags=re.DOTALL)
    html = re.sub(r"<code[^>]*>.*?</code>", _stash, html, flags=re.DOTALL)

    def _replace(match: re.Match[str]) -> str:
        status_text = match.group(1).strip().upper()
        color_map = {
            "ON TRACK": "Green",
            "AT RISK": "Yellow",
            "OFF TRACK": "Red",
            "BLOCKED": "Red",
            "DONE": "Blue",
            "DRAFT": "Grey",
        }
        color = color_map.get(status_text, "Grey")
        return (
            '<ac:structured-macro ac:name="status">'
            f'<ac:parameter ac:name="title">{status_text}</ac:parameter>'
            f'<ac:parameter ac:name="colour">{color}</ac:parameter>'
            "</ac:structured-macro>"
        )

    html = re.sub(r"\[STATUS:\s*([A-Za-z0-9 _-]+)\]", _replace, html)

    for idx, block in enumerate(placeholders):
        html = html.replace(f"__GG_CODE_BLOCK_{idx}__", block)
    return html


def _inject_toc(html: str) -> str:
    if 'ac:name="toc"' in html:
        return html
    header_count = len(re.findall(r"<h[1-3]\b", html))
    if header_count >= 3:
        toc_macro = '<p><ac:structured-macro ac:name="toc"/></p>'
        return f"{toc_macro}{html}"
    return html


def _apply_confluence_style(html: str) -> str:
    html = _transform_container_divs(html)
    html = _transform_blockquotes(html)
    html = _transform_status_lozenges(html)
    html = _inject_toc(html)
    return html


def markdown_to_storage(markdown: str, external_id: Optional[str] = None) -> str:
    md = _build_markdown_renderer()
    html = md.render(markdown or "")
    html = _apply_confluence_style(html)
    html = _xhtml_clean(html)

    if external_id:
        marker = f"{CONFLUENCE_MARKER_PREFIX}{external_id}"
        marker_comment = f"<!-- {marker} -->"
        marker_block = f'{marker_comment}\n<p style="display:none" data-g-gremlin="external-id">{marker}</p>'
        if html.strip():
            html = f"{html.rstrip()}\n{marker_block}\n"
        else:
            html = f"{marker_block}\n"
    return html


class NotionClient(DocsSinkProvider):
    """Notion provider implementation with schema-aware writes and read helpers."""

    def __init__(self, token: str, database_id: Optional[str] = None) -> None:
        self.token = token
        self.database_id = database_id
        self.session = requests.Session()
        self._database_schema_cache: Dict[str, Dict[str, Any]] = {}

    @property
    def name(self) -> str:  # pragma: no cover - trivial
        return "notion"

    @property
    def provider_type(self) -> DocsProvider:  # pragma: no cover - trivial
        return DocsProvider.NOTION

    def _headers(self) -> Dict[str, str]:
        return {
            "Authorization": f"Bearer {self.token}",
            "Notion-Version": NOTION_VERSION,
            "Content-Type": "application/json",
        }

    def _request(
        self,
        method: str,
        path: str,
        json: Optional[Dict[str, Any]] = None,
        params: Optional[Dict[str, Any]] = None,
    ) -> requests.Response:
        url = f"{NOTION_API_BASE}{path}"
        resp = self.session.request(
            method,
            url,
            headers=self._headers(),
            json=json,
            params=params,
            timeout=30,
        )
        if resp.status_code == 429:
            retry_after = int(resp.headers.get("Retry-After", "1"))
            time.sleep(retry_after)
            resp = self.session.request(
                method,
                url,
                headers=self._headers(),
                json=json,
                params=params,
                timeout=30,
            )
        return resp

    def test_connection(self) -> Dict[str, Any]:
        resp = self._request("GET", "/users/me")
        if resp.status_code == 401:
            raise DocsSinkError("Notion auth failed")
        if not resp.ok:
            raise DocsSinkError(f"Notion connection failed ({resp.status_code})")
        data = resp.json()
        return {"status": "ok", "bot": data.get("bot"), "user": data.get("user")}

    def get_database(self, database_id: Optional[str] = None) -> Dict[str, Any]:
        resolved_database_id = database_id or self.database_id
        if not resolved_database_id:
            raise DocsSinkError("Notion database_id is required.")
        resp = self._request("GET", f"/databases/{resolved_database_id}")
        if not resp.ok:
            raise DocsSinkError(f"Notion database lookup failed ({resp.status_code})")
        return resp.json()

    def get_database_schema(self, database_id: Optional[str] = None) -> Dict[str, Any]:
        resolved_database_id = database_id or self.database_id
        if not resolved_database_id:
            raise DocsSinkError("Notion database_id is required.")
        cached = self._database_schema_cache.get(resolved_database_id)
        if cached is not None:
            return cached

        database = self.get_database(resolved_database_id)
        properties = database.get("properties") or {}
        schema = {
            "id": database.get("id", resolved_database_id),
            "title": self._plain_text_from_rich_text(database.get("title") or []),
            "properties": properties,
            "title_property": self._find_property_by_type(properties, "title"),
            "external_id_property": self._find_external_id_property(properties),
            "tags_property": self._find_tags_property(properties),
        }
        self._database_schema_cache[resolved_database_id] = schema
        return schema

    def list_databases(self, query: Optional[str] = None, limit: int = 25) -> List[Dict[str, Any]]:
        results: List[Dict[str, Any]] = []
        start_cursor: Optional[str] = None

        while len(results) < limit:
            payload: Dict[str, Any] = {
                "filter": {"property": "object", "value": "database"},
                "page_size": min(100, limit - len(results)),
            }
            if query:
                payload["query"] = query
            if start_cursor:
                payload["start_cursor"] = start_cursor

            resp = self._request("POST", "/search", json=payload)
            if not resp.ok:
                raise DocsSinkError(f"Notion database search failed ({resp.status_code})")

            data = resp.json()
            results.extend(data.get("results", []))
            if not data.get("has_more") or not data.get("next_cursor"):
                break
            start_cursor = data.get("next_cursor")

        return [self._simplify_database(database) for database in results[:limit]]

    def _get_page_by_external_id_in_database(
        self,
        external_id: str,
        database_id: Optional[str],
    ) -> Optional[Dict[str, Any]]:
        if not database_id:
            return None
        schema = self.get_database_schema(database_id)
        external_id_property = schema.get("external_id_property")
        if not external_id_property:
            raise DocsSinkError(
                "Notion database must include an 'external_id' style property for idempotent upserts."
            )

        properties = schema.get("properties") or {}
        payload = {
            "filter": self._build_external_id_filter(
                str(external_id_property),
                properties.get(str(external_id_property), {}),
                external_id,
            ),
            "page_size": 1,
        }
        resp = self._request("POST", f"/databases/{database_id}/query", json=payload)
        if not resp.ok:
            return None
        results = resp.json().get("results", [])
        return results[0] if results else None

    def get_page_by_external_id(self, external_id: str) -> Optional[Dict[str, Any]]:
        return self._get_page_by_external_id_in_database(external_id, self.database_id)

    def upsert_page(self, page: DocsPage) -> DocsPublishResult:
        database_id = page.parent_id or self.database_id
        if not database_id:
            raise DocsSinkError("Notion database_id is required.")

        schema = self.get_database_schema(database_id)
        existing = self._get_page_by_external_id_in_database(page.external_id, database_id)
        properties = self._build_properties(page, schema=schema)
        if existing:
            page_id = existing.get("id")
            update_payload = {
                "properties": properties,
            }
            resp = self._request("PATCH", f"/pages/{page_id}", json=update_payload)
            if not resp.ok:
                return DocsPublishResult(
                    external_id=page.external_id,
                    status="failed",
                    page_id=page_id,
                    page_url=None,
                    error=f"HTTP {resp.status_code}",
                )
            try:
                self._replace_page_blocks(str(page_id), page.content_markdown)
            except DocsSinkError as exc:
                return DocsPublishResult(
                    external_id=page.external_id,
                    status="failed",
                    page_id=page_id,
                    page_url=existing.get("url"),
                    error=str(exc),
                )
            data = resp.json() if resp.content else {}
            url = data.get("url") or existing.get("url")
            return DocsPublishResult(
                external_id=page.external_id,
                status="updated",
                page_id=page_id,
                page_url=url,
            )

        blocks = self._markdown_to_blocks(page.content_markdown)
        create_payload = {
            "parent": {"database_id": database_id},
            "properties": properties,
        }
        if blocks:
            create_payload["children"] = blocks[:100]
        resp = self._request("POST", "/pages", json=create_payload)
        if not resp.ok:
            return DocsPublishResult(
                external_id=page.external_id,
                status="failed",
                page_id=None,
                page_url=None,
                error=f"HTTP {resp.status_code}",
            )
        data = resp.json()
        page_id = data.get("id")
        if page_id and len(blocks) > 100:
            try:
                self._append_block_children(str(page_id), blocks[100:])
            except DocsSinkError as exc:
                return DocsPublishResult(
                    external_id=page.external_id,
                    status="failed",
                    page_id=page_id,
                    page_url=data.get("url"),
                    error=str(exc),
                )
        return DocsPublishResult(
            external_id=page.external_id,
            status="created",
            page_id=page_id,
            page_url=data.get("url"),
        )

    def upsert_pages_batch(self, pages: List[DocsPage]) -> BatchOutcome:
        outcome = BatchOutcome()
        for page in pages:
            result = self.upsert_page(page)
            if result.status in ("created", "updated"):
                outcome.succeeded.append(result.to_dict())
            elif result.status == "failed":
                outcome.failed.append(result.to_dict())
            else:
                outcome.retryable_failed.append(result.to_dict())
        return outcome

    def delete_page(self, page_id: str) -> bool:
        resp = self._request("PATCH", f"/pages/{page_id}", json={"archived": True})
        return resp.ok

    def get_page(self, page_id: str) -> Dict[str, Any]:
        resp = self._request("GET", f"/pages/{page_id}")
        if not resp.ok:
            raise DocsSinkError(f"Notion page lookup failed ({resp.status_code})")
        return resp.json()

    def query_database_pages(self, database_id: Optional[str] = None, limit: int = 25) -> List[Dict[str, Any]]:
        resolved_database_id = database_id or self.database_id
        if not resolved_database_id:
            raise DocsSinkError("Notion database_id is required.")

        results: List[Dict[str, Any]] = []
        start_cursor: Optional[str] = None

        while len(results) < limit:
            payload: Dict[str, Any] = {
                "page_size": min(100, limit - len(results)),
            }
            if start_cursor:
                payload["start_cursor"] = start_cursor

            resp = self._request("POST", f"/databases/{resolved_database_id}/query", json=payload)
            if not resp.ok:
                raise DocsSinkError(f"Notion database query failed ({resp.status_code})")

            data = resp.json()
            results.extend(data.get("results", []))
            if not data.get("has_more") or not data.get("next_cursor"):
                break
            start_cursor = data.get("next_cursor")

        return results[:limit]

    def search_pages(
        self,
        query: str,
        database_id: Optional[str] = None,
        limit: int = 25,
    ) -> List[Dict[str, Any]]:
        results: List[Dict[str, Any]] = []
        start_cursor: Optional[str] = None

        while len(results) < limit:
            payload: Dict[str, Any] = {
                "query": query,
                "filter": {"property": "object", "value": "page"},
                "page_size": min(100, limit - len(results)),
            }
            if start_cursor:
                payload["start_cursor"] = start_cursor

            resp = self._request("POST", "/search", json=payload)
            if not resp.ok:
                raise DocsSinkError(f"Notion page search failed ({resp.status_code})")

            data = resp.json()
            page_results = data.get("results", [])
            if database_id:
                page_results = [
                    item for item in page_results if self._page_belongs_to_database(item, database_id)
                ]
            results.extend(page_results)
            if not data.get("has_more") or not data.get("next_cursor"):
                break
            start_cursor = data.get("next_cursor")

        return results[:limit]

    def get_page_content(self, page_id: str) -> List[Dict[str, Any]]:
        return self._populate_block_children(self._list_block_children(page_id))

    def simplify_page(self, page: Mapping[str, Any]) -> Dict[str, Any]:
        parent = page.get("parent") or {}
        return {
            "id": page.get("id"),
            "url": page.get("url"),
            "title": self._extract_page_title(page),
            "database_id": parent.get("database_id") or parent.get("data_source_id"),
            "created_time": page.get("created_time"),
            "last_edited_time": page.get("last_edited_time"),
            "properties": self._simplify_page_properties(page),
        }

    def blocks_to_plain_text(self, blocks: Iterable[Mapping[str, Any]]) -> str:
        lines: List[str] = []
        for block in blocks:
            lines.extend(self._block_to_plain_text_lines(block))
        text = "\n".join(line for line in lines if line is not None).strip()
        return f"{text}\n" if text else ""

    def _build_properties(self, page: DocsPage, schema: Optional[Mapping[str, Any]] = None) -> Dict[str, Any]:
        if schema is None:
            return self._build_legacy_properties(page)

        schema_properties = schema.get("properties") or {}
        title_property = schema.get("title_property")
        external_id_property = schema.get("external_id_property")
        if not title_property:
            raise DocsSinkError("Notion database is missing a title property.")
        if not external_id_property:
            raise DocsSinkError(
                "Notion database must include an 'external_id' style property for idempotent upserts."
            )

        props: Dict[str, Any] = {}
        matched: set[str] = set()

        title_payload = self._property_value_payload(page.title or "Untitled", schema_properties.get(title_property, {}))
        if title_payload is not None:
            props[title_property] = title_payload
            matched.add(str(title_property))

        external_id_payload = self._property_value_payload(page.external_id, schema_properties.get(external_id_property, {}))
        if external_id_payload is not None:
            props[external_id_property] = external_id_payload
            matched.add(str(external_id_property))

        for key, value in page.properties.items():
            property_name = self._match_schema_property_name(key, schema_properties)
            if not property_name or property_name in matched:
                continue
            payload = self._property_value_payload(value, schema_properties.get(property_name, {}))
            if payload is not None:
                props[property_name] = payload
                matched.add(property_name)

        tags_property = schema.get("tags_property")
        if tags_property and page.tags and tags_property not in matched:
            property_meta = schema_properties.get(tags_property, {})
            raw_value: Any = page.tags
            if property_meta.get("type") in {"select", "status"}:
                raw_value = page.tags[0]
            payload = self._property_value_payload(raw_value, property_meta)
            if payload is not None:
                props[tags_property] = payload

        return props

    def _build_legacy_properties(self, page: DocsPage) -> Dict[str, Any]:
        props: Dict[str, Any] = {
            "Name": {
                "title": self._rich_text_value(page.title or "Untitled"),
            },
            "external_id": {
                "rich_text": self._rich_text_value(page.external_id),
            },
        }
        for key, value in page.properties.items():
            props[key] = {"rich_text": self._rich_text_value(str(value))}
        if page.tags:
            props["Tags"] = {"multi_select": [{"name": tag} for tag in page.tags]}
        return props

    def _markdown_to_blocks(self, markdown: str) -> List[Dict[str, Any]]:
        if not markdown:
            return []
        blocks: List[Dict[str, Any]] = []
        in_code_block = False
        code_language = "plain text"
        code_lines: List[str] = []

        for line in markdown.splitlines():
            content = line.strip()

            if content.startswith("```"):
                if in_code_block:
                    blocks.extend(self._build_code_blocks("\n".join(code_lines), code_language))
                    in_code_block = False
                    code_language = "plain text"
                    code_lines = []
                else:
                    in_code_block = True
                    code_language = self._normalize_notion_code_language(content[3:].strip())
                    code_lines = []
                continue

            if in_code_block:
                code_lines.append(line)
                continue

            if not content:
                continue

            block_type, block_content = self._line_to_notion_block(content)
            blocks.append(
                {
                    "object": "block",
                    "type": block_type,
                    block_type: {
                        "rich_text": [{"type": "text", "text": {"content": block_content[:2000]}}],
                    },
                }
            )

        if in_code_block:
            blocks.extend(self._build_code_blocks("\n".join(code_lines), code_language))

        return blocks

    def _list_block_children(self, block_id: str) -> List[Dict[str, Any]]:
        blocks: List[Dict[str, Any]] = []
        start_cursor: Optional[str] = None

        while True:
            params: Dict[str, Any] = {"page_size": 100}
            if start_cursor:
                params["start_cursor"] = start_cursor

            resp = self._request("GET", f"/blocks/{block_id}/children", params=params)
            if not resp.ok:
                raise DocsSinkError(f"Notion block children lookup failed ({resp.status_code})")

            data = resp.json()
            blocks.extend(data.get("results", []))
            if not data.get("has_more") or not data.get("next_cursor"):
                break
            start_cursor = data.get("next_cursor")

        return blocks

    def _populate_block_children(self, blocks: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        populated: List[Dict[str, Any]] = []
        for block in blocks:
            current = dict(block)
            if current.get("has_children") and current.get("id"):
                current["children"] = self._populate_block_children(
                    self._list_block_children(str(current["id"]))
                )
            populated.append(current)
        return populated

    def _replace_page_blocks(self, page_id: str, markdown: str) -> None:
        existing_blocks = self._list_block_children(page_id)
        if len(existing_blocks) >= NOTION_LARGE_REPLACE_WARNING_BLOCKS:
            LOGGER.warning(
                "Replacing large Notion page body with %s existing blocks for page %s; "
                "updates may be slow because Notion requires per-block deletes.",
                len(existing_blocks),
                page_id,
            )

        for block in existing_blocks:
            block_id = block.get("id")
            if not block_id:
                continue
            resp = self._request("DELETE", f"/blocks/{block_id}")
            if not resp.ok:
                raise DocsSinkError(
                    f"Notion block delete failed ({resp.status_code}). "
                    "Page content may be partially deleted; re-run publish to repair."
                )
        self._append_block_children(page_id, self._markdown_to_blocks(markdown))

    def _append_block_children(self, block_id: str, blocks: List[Dict[str, Any]]) -> None:
        for start in range(0, len(blocks), 100):
            resp = self._request(
                "PATCH",
                f"/blocks/{block_id}/children",
                json={"children": blocks[start:start + 100]},
            )
            if not resp.ok:
                raise DocsSinkError(
                    f"Notion block append failed ({resp.status_code}). "
                    "Page content replacement may be incomplete; re-run publish to repair."
                )

    def _line_to_notion_block(self, content: str) -> Tuple[str, str]:
        if content.startswith("# "):
            return "heading_1", content[2:].strip() or " "
        if content.startswith("## "):
            return "heading_2", content[3:].strip() or " "
        if content.startswith("### "):
            return "heading_3", content[4:].strip() or " "
        if re.match(r"^\d+\.\s+", content):
            stripped = re.sub(r"^\d+\.\s+", "", content).strip()
            return "numbered_list_item", stripped or " "
        if content.startswith("- ") or content.startswith("* "):
            return "bulleted_list_item", content[2:].strip() or " "
        return "paragraph", content

    def _normalize_notion_code_language(self, language: str) -> str:
        if not language:
            return "plain text"
        normalized = language.lower()
        mapping = {
            "py": "python",
            "js": "javascript",
            "ts": "typescript",
            "sh": "bash",
            "shell": "bash",
            "md": "markdown",
            "mmd": "plain text",
            "mermaid": "plain text",
            "text": "plain text",
            "plaintext": "plain text",
        }
        notion_language = mapping.get(normalized, normalized)
        allowed = {
            "abap", "agda", "arduino", "ascii art", "assembly", "bash", "basic", "bnf", "c", "c#", "c++",
            "clojure", "coffeescript", "coq", "css", "dart", "dhall", "diff", "docker", "ebnf", "elixir", "elm",
            "erlang", "f#", "flow", "fortran", "gherkin", "glsl", "go", "graphql", "groovy", "haskell", "html",
            "idris", "java", "javascript", "json", "julia", "kotlin", "latex", "less", "lisp", "livescript", "lua",
            "makefile", "markdown", "markup", "matlab", "mermaid", "nix", "notion formula", "objective-c",
            "ocaml", "pascal", "perl", "php", "plain text", "powershell", "prolog", "protobuf", "purescript",
            "python", "r", "reason", "ruby", "rust", "sass", "scala", "scheme", "scss", "shell", "solidity",
            "sql", "swift", "toml", "typescript", "vb.net", "verilog", "vhdl", "visual basic", "webassembly",
            "xml", "yaml", "java/c/c++/c#",
        }
        return notion_language if notion_language in allowed else "plain text"

    def _build_external_id_filter(
        self,
        property_name: str,
        property_meta: Mapping[str, Any],
        external_id: str,
    ) -> Dict[str, Any]:
        property_type = property_meta.get("type", "rich_text")
        if property_type == "title":
            return {"property": property_name, "title": {"equals": external_id}}
        if property_type == "number":
            numeric_external_id = self._coerce_number(external_id)
            if numeric_external_id is None:
                raise DocsSinkError("Notion external_id property is numeric, but the supplied external_id is not.")
            return {"property": property_name, "number": {"equals": numeric_external_id}}
        return {"property": property_name, "rich_text": {"equals": external_id}}

    def _property_value_payload(
        self,
        value: Any,
        property_meta: Mapping[str, Any],
    ) -> Optional[Dict[str, Any]]:
        property_type = property_meta.get("type")
        if property_type == "title":
            return {"title": self._rich_text_value(value or "Untitled")}
        if property_type == "rich_text":
            return {"rich_text": self._rich_text_value(value)}
        if property_type == "number":
            return {"number": self._coerce_number(value)}
        if property_type == "checkbox":
            return {"checkbox": self._coerce_bool(value)}
        if property_type == "select":
            selected = self._coerce_optional_string(value)
            return {"select": {"name": selected} if selected else None}
        if property_type == "multi_select":
            return {"multi_select": [{"name": item} for item in self._coerce_string_list(value)]}
        if property_type == "status":
            selected = self._coerce_optional_string(value)
            return {"status": {"name": selected} if selected else None}
        if property_type == "url":
            return {"url": self._coerce_optional_string(value)}
        if property_type == "email":
            return {"email": self._coerce_optional_string(value)}
        if property_type == "phone_number":
            return {"phone_number": self._coerce_optional_string(value)}
        if property_type == "date":
            return {"date": self._coerce_date_value(value)}
        if property_type == "relation":
            return {"relation": [{"id": item} for item in self._coerce_id_list(value)]}
        if property_type == "people":
            return {"people": [{"id": item} for item in self._coerce_id_list(value)]}
        if property_type in {
            "formula",
            "rollup",
            "created_by",
            "created_time",
            "last_edited_by",
            "last_edited_time",
            "unique_id",
            "verification",
            "button",
            "files",
        }:
            return None
        return {"rich_text": self._rich_text_value(value)}

    def _rich_text_value(self, value: Any) -> List[Dict[str, Any]]:
        text = self._coerce_optional_string(value)
        if not text:
            return []
        return [
            {"type": "text", "text": {"content": text[i:i + 2000]}}
            for i in range(0, len(text), 2000)
        ]

    def _coerce_optional_string(self, value: Any) -> Optional[str]:
        if value is None:
            return None
        text = str(value).strip()
        return text or None

    def _coerce_number(self, value: Any) -> Optional[float]:
        if value is None:
            return None
        if isinstance(value, bool):
            return 1.0 if value else 0.0
        if isinstance(value, (int, float)):
            return float(value)
        text = str(value).strip().replace(",", "")
        if not text:
            return None
        try:
            return float(text)
        except ValueError:
            return None

    def _coerce_bool(self, value: Any) -> bool:
        if isinstance(value, bool):
            return value
        if isinstance(value, (int, float)):
            return bool(value)
        text = str(value).strip().lower()
        return text in {"1", "true", "yes", "y", "on", "checked"}

    def _coerce_string_list(self, value: Any) -> List[str]:
        if value is None:
            return []
        if isinstance(value, str):
            parts = [part.strip() for part in re.split(r"[;,]", value) if part.strip()]
            return parts or ([value.strip()] if value.strip() else [])
        if isinstance(value, (list, tuple, set)):
            items: List[str] = []
            for item in value:
                text = self._coerce_optional_string(item)
                if text:
                    items.append(text)
            return items
        text = self._coerce_optional_string(value)
        return [text] if text else []

    def _coerce_id_list(self, value: Any) -> List[str]:
        if isinstance(value, (list, tuple, set)):
            raw_values: Iterable[Any] = value
        else:
            raw_values = self._coerce_string_list(value)
        items: List[str] = []
        for item in raw_values:
            candidate = item.get("id") if isinstance(item, Mapping) else item
            text = self._coerce_optional_string(candidate)
            if text:
                items.append(text)
        return items

    def _coerce_date_value(self, value: Any) -> Optional[Dict[str, Any]]:
        if value is None:
            return None
        if isinstance(value, Mapping):
            start = value.get("start") or value.get("date")
            if not start:
                return None
            payload: Dict[str, Any] = {"start": str(start)}
            if value.get("end"):
                payload["end"] = str(value["end"])
            if value.get("time_zone"):
                payload["time_zone"] = str(value["time_zone"])
            return payload
        text = self._coerce_optional_string(value)
        return {"start": text} if text else None

    def _simplify_database(self, database: Mapping[str, Any]) -> Dict[str, Any]:
        return {
            "id": database.get("id"),
            "title": self._plain_text_from_rich_text(database.get("title") or []),
            "url": database.get("url"),
            "last_edited_time": database.get("last_edited_time"),
            "property_count": len(database.get("properties") or {}),
        }

    def _simplify_page_properties(self, page: Mapping[str, Any]) -> Dict[str, Any]:
        simplified: Dict[str, Any] = {}
        for name, property_value in (page.get("properties") or {}).items():
            if isinstance(property_value, Mapping):
                simplified[name] = self._plain_text_from_property(property_value)
        return simplified

    def _plain_text_from_property(self, property_value: Mapping[str, Any]) -> Any:
        property_type = property_value.get("type")
        if property_type == "title":
            return self._plain_text_from_rich_text(property_value.get("title") or [])
        if property_type == "rich_text":
            return self._plain_text_from_rich_text(property_value.get("rich_text") or [])
        if property_type == "number":
            return property_value.get("number")
        if property_type == "checkbox":
            return property_value.get("checkbox")
        if property_type in {"select", "status"}:
            selected = property_value.get(property_type) or {}
            return selected.get("name")
        if property_type == "multi_select":
            return [item.get("name") for item in property_value.get("multi_select") or [] if item.get("name")]
        if property_type in {"url", "email", "phone_number"}:
            return property_value.get(property_type)
        if property_type == "date":
            return property_value.get("date")
        if property_type in {"relation", "people"}:
            return [item.get("id") for item in property_value.get(property_type) or [] if item.get("id")]
        if property_type in {"created_time", "last_edited_time"}:
            return property_value.get(property_type)
        if property_type in {"created_by", "last_edited_by"}:
            actor = property_value.get(property_type) or {}
            return actor.get("id")
        if property_type == "formula":
            formula = property_value.get("formula") or {}
            return formula.get(formula.get("type"))
        return property_value.get(property_type)

    def _extract_page_title(self, page: Mapping[str, Any]) -> str:
        for property_value in (page.get("properties") or {}).values():
            if isinstance(property_value, Mapping) and property_value.get("type") == "title":
                return self._plain_text_from_rich_text(property_value.get("title") or []) or "Untitled"
        return "Untitled"

    def _plain_text_from_rich_text(self, rich_text: Iterable[Mapping[str, Any]]) -> str:
        parts: List[str] = []
        for item in rich_text:
            plain_text = item.get("plain_text")
            if plain_text:
                parts.append(str(plain_text))
                continue
            text = item.get("text") or {}
            content = text.get("content")
            if content:
                parts.append(str(content))
        return "".join(parts)

    def _block_to_plain_text_lines(self, block: Mapping[str, Any], indent: int = 0) -> List[str]:
        block_type = str(block.get("type") or "")
        block_text = self._plain_text_from_block(block)
        prefix = " " * indent
        lines: List[str]

        if block_type == "heading_1":
            lines = [f"{prefix}# {block_text}".rstrip()]
        elif block_type == "heading_2":
            lines = [f"{prefix}## {block_text}".rstrip()]
        elif block_type == "heading_3":
            lines = [f"{prefix}### {block_text}".rstrip()]
        elif block_type == "bulleted_list_item":
            lines = [f"{prefix}- {block_text}".rstrip()]
        elif block_type == "numbered_list_item":
            lines = [f"{prefix}1. {block_text}".rstrip()]
        elif block_type == "quote":
            lines = [f"{prefix}> {block_text}".rstrip()]
        elif block_type == "to_do":
            checked = bool((block.get("to_do") or {}).get("checked"))
            marker = "x" if checked else " "
            lines = [f"{prefix}- [{marker}] {block_text}".rstrip()]
        elif block_type == "code":
            language = ((block.get("code") or {}).get("language") or "").strip()
            code_lines = block_text.splitlines() or [""]
            lines = [f"{prefix}```{language}".rstrip()]
            lines.extend(f"{prefix}{line}" for line in code_lines)
            lines.append(f"{prefix}```")
        else:
            lines = [f"{prefix}{block_text}".rstrip()] if block_text else []

        child_indent = indent + (2 if block_type in {"bulleted_list_item", "numbered_list_item", "to_do"} else 0)
        for child in block.get("children") or []:
            lines.extend(self._block_to_plain_text_lines(child, child_indent))
        return lines

    def _plain_text_from_block(self, block: Mapping[str, Any]) -> str:
        block_type = str(block.get("type") or "")
        block_data = block.get(block_type) or {}
        if not isinstance(block_data, Mapping):
            return ""
        if block_type == "child_page":
            return str(block_data.get("title") or "")
        if block_type == "table_row":
            return " | ".join(
                self._plain_text_from_rich_text(cell)
                for cell in block_data.get("cells") or []
            )
        return self._plain_text_from_rich_text(block_data.get("rich_text") or [])

    def _page_belongs_to_database(self, page: Mapping[str, Any], database_id: str) -> bool:
        parent = page.get("parent") or {}
        return parent.get("database_id") == database_id or parent.get("data_source_id") == database_id

    def _find_property_by_type(self, properties: Mapping[str, Any], property_type: str) -> Optional[str]:
        for name, meta in properties.items():
            if isinstance(meta, Mapping) and meta.get("type") == property_type:
                return name
        return None

    def _find_external_id_property(self, properties: Mapping[str, Any]) -> Optional[str]:
        for candidate in ("external_id", "External ID", "external-id", "external id"):
            matched = self._match_schema_property_name(candidate, properties)
            if matched:
                return matched
        return None

    def _find_tags_property(self, properties: Mapping[str, Any]) -> Optional[str]:
        for candidate in ("tags", "Tags", "labels", "Labels"):
            matched = self._match_schema_property_name(candidate, properties)
            if not matched:
                continue
            property_type = (properties.get(matched) or {}).get("type")
            if property_type in {"multi_select", "select", "status"}:
                return matched
        return None

    def _match_schema_property_name(
        self,
        key: str,
        properties: Mapping[str, Any],
    ) -> Optional[str]:
        if key in properties:
            return key
        normalized_key = self._normalize_property_name(key)
        for name in properties:
            if self._normalize_property_name(name) == normalized_key:
                return name
        return None

    def _normalize_property_name(self, value: str) -> str:
        return re.sub(r"[^a-z0-9]+", "", value.lower())

    def _build_code_blocks(self, content: str, language: str) -> List[Dict[str, Any]]:
        text = content or " "
        chunks = [text[i:i + 2000] for i in range(0, len(text), 2000)] or [" "]
        blocks: List[Dict[str, Any]] = []
        for chunk in chunks:
            blocks.append(
                {
                    "object": "block",
                    "type": "code",
                    "code": {
                        "rich_text": [{"type": "text", "text": {"content": chunk}}],
                        "language": language,
                    },
                }
            )
        return blocks


class OutlineClient(DocsSinkProvider):
    """Outline provider implementation."""

    def __init__(
        self,
        api_token: str,
        api_base: Optional[str] = None,
        default_collection_id: Optional[str] = None,
    ) -> None:
        self.api_token = api_token
        self.api_base = self._normalize_base(api_base or OUTLINE_DEFAULT_API_BASE)
        self.default_collection_id = default_collection_id
        self.session = requests.Session()

    @property
    def name(self) -> str:  # pragma: no cover - trivial
        return "outline"

    @property
    def provider_type(self) -> DocsProvider:  # pragma: no cover - trivial
        return DocsProvider.OUTLINE

    def _normalize_base(self, api_base: str) -> str:
        base = api_base.rstrip("/")
        if not base.endswith("/api"):
            base = f"{base}/api"
        return base

    def _headers(self) -> Dict[str, str]:
        return {
            "Authorization": f"Bearer {self.api_token}",
            "Content-Type": "application/json",
            "Accept": "application/json",
        }

    def _post(self, method: str, payload: Optional[Dict[str, Any]] = None) -> requests.Response:
        path = method if method.startswith("/") else f"/{method}"
        url = f"{self.api_base}{path}"
        resp = self.session.post(url, headers=self._headers(), json=payload or {}, timeout=30)
        if resp.status_code == 429:
            retry_after = int(resp.headers.get("Retry-After", "1"))
            time.sleep(retry_after)
            resp = self.session.post(url, headers=self._headers(), json=payload or {}, timeout=30)
        return resp

    def _response_ok(self, resp: requests.Response) -> bool:
        if not resp.ok:
            return False
        try:
            data = resp.json()
            return bool(data.get("ok", True))
        except Exception:
            return False

    def test_connection(self) -> Dict[str, Any]:
        resp = self._post("/auth.info")
        if resp.status_code == 401:
            raise DocsSinkError("Outline auth failed")
        if not self._response_ok(resp):
            raise DocsSinkError(f"Outline connection failed ({resp.status_code})")
        data = resp.json()
        return {"status": "ok", "team": data.get("data", {}).get("team"), "user": data.get("data", {}).get("user")}

    def list_collections(self) -> List[Dict[str, Any]]:
        """Return all collections from the Outline instance."""
        results: List[Dict[str, Any]] = []
        offset = 0
        limit = 25
        while True:
            resp = self._post("/collections.list", {"offset": offset, "limit": limit})
            if not self._response_ok(resp):
                raise DocsSinkError(f"Outline collections.list failed ({resp.status_code})")
            data = resp.json().get("data", [])
            if not data:
                break
            for col in data:
                results.append({
                    "id": col.get("id"),
                    "name": col.get("name"),
                    "urlId": col.get("urlId"),
                    "description": col.get("description", ""),
                    "documents_count": col.get("documents", {}).get("count") if isinstance(col.get("documents"), dict) else None,
                })
            if len(data) < limit:
                break
            offset += limit
        return results

    def resolve_collection_id(self, id_or_slug: str) -> str:
        """Resolve a collection UUID or URL slug to the API-level UUID.

        Accepts either a full UUID or the short ``urlId`` visible in Outline
        URLs (e.g. ``qbT0rAgCPW``).  Returns the UUID required by the API.
        """
        import re

        _UUID_RE = re.compile(
            r"^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$", re.IGNORECASE
        )

        if _UUID_RE.match(id_or_slug):
            return id_or_slug

        collections = self.list_collections()
        for col in collections:
            if col.get("urlId") == id_or_slug or col.get("name", "").lower() == id_or_slug.lower():
                return col["id"]

        raise DocsSinkError(
            f"Could not resolve Outline collection '{id_or_slug}'. "
            "Use a UUID or a valid urlId/name from 'g-gremlin docs collections --provider outline'."
        )

    def get_page_by_external_id(self, external_id: str) -> Optional[Dict[str, Any]]:
        payload: Dict[str, Any] = {"query": external_id}
        if self.default_collection_id:
            payload["collectionId"] = self.default_collection_id
        resp = self._post("/documents.search", payload)
        if not self._response_ok(resp):
            return None
        results = resp.json().get("data", [])
        for item in results:
            document = item.get("document") or {}
            text = document.get("text") or ""
            if self._matches_external_id(text, external_id):
                return document
        return None

    def upsert_page(self, page: DocsPage) -> DocsPublishResult:
        collection_id = page.parent_id or self.default_collection_id
        if not collection_id:
            raise DocsSinkError("Outline collection_id is required.")

        collection_id = self.resolve_collection_id(collection_id)
        content = self._ensure_metadata(page, collection_id)
        existing = self.get_page_by_external_id(page.external_id)

        if existing:
            document_id = existing.get("id")
            payload = {
                "id": document_id,
                "title": page.title,
                "text": content,
                "publish": True,
            }
            resp = self._post("/documents.update", payload)
            if not self._response_ok(resp):
                return DocsPublishResult(
                    external_id=page.external_id,
                    status="failed",
                    page_id=document_id,
                    page_url=None,
                    error=f"HTTP {resp.status_code}",
                )
            data = resp.json().get("data", existing)
            return DocsPublishResult(
                external_id=page.external_id,
                status="updated",
                page_id=data.get("id"),
                page_url=self._document_url(data),
            )

        payload = {
            "collectionId": collection_id,
            "title": page.title,
            "text": content,
            "publish": True,
        }
        resp = self._post("/documents.create", payload)
        if not self._response_ok(resp):
            return DocsPublishResult(
                external_id=page.external_id,
                status="failed",
                page_id=None,
                page_url=None,
                error=f"HTTP {resp.status_code}",
            )
        data = resp.json().get("data", {})
        return DocsPublishResult(
            external_id=page.external_id,
            status="created",
            page_id=data.get("id"),
            page_url=self._document_url(data),
        )

    def upsert_pages_batch(self, pages: List[DocsPage]) -> BatchOutcome:
        outcome = BatchOutcome()
        for page in pages:
            result = self.upsert_page(page)
            if result.status in ("created", "updated"):
                outcome.succeeded.append(result.to_dict())
            elif result.status == "failed":
                outcome.failed.append(result.to_dict())
            else:
                outcome.retryable_failed.append(result.to_dict())
        return outcome

    def delete_page(self, page_id: str) -> bool:
        resp = self._post("/documents.delete", {"id": page_id})
        return self._response_ok(resp)

    # ------------------------------------------------------------------
    # Read-only helpers
    # ------------------------------------------------------------------

    def get_document(self, doc_id: str) -> Dict[str, Any]:
        """Fetch a single document by ID (UUID or urlId slug)."""
        resp = self._post("/documents.info", {"id": doc_id})
        if resp.status_code == 404:
            raise DocsSinkError(f"Document '{doc_id}' not found.")
        if not self._response_ok(resp):
            raise DocsSinkError(
                f"Outline documents.info failed ({resp.status_code})"
            )
        return resp.json().get("data", {})

    def search_documents(
        self,
        query: str,
        collection_id: Optional[str] = None,
        limit: int = 25,
    ) -> List[Dict[str, Any]]:
        """Full-text search via /documents.search."""
        payload: Dict[str, Any] = {"query": query, "limit": min(limit, 25)}
        effective = collection_id or self.default_collection_id
        if effective:
            payload["collectionId"] = effective
        resp = self._post("/documents.search", payload)
        if not self._response_ok(resp):
            raise DocsSinkError(
                f"Outline documents.search failed ({resp.status_code})"
            )
        return resp.json().get("data", [])

    def list_documents(
        self,
        collection_id: Optional[str] = None,
        limit: int = 25,
        offset: int = 0,
    ) -> List[Dict[str, Any]]:
        """List documents via /documents.list."""
        payload: Dict[str, Any] = {"offset": offset, "limit": limit}
        effective = collection_id or self.default_collection_id
        if effective:
            payload["collectionId"] = effective
        resp = self._post("/documents.list", payload)
        if not self._response_ok(resp):
            raise DocsSinkError(
                f"Outline documents.list failed ({resp.status_code})"
            )
        return resp.json().get("data", [])

    def _matches_external_id(self, text: str, external_id: str) -> bool:
        marker = f"{OUTLINE_EXTERNAL_ID_MARKER} {external_id}"
        return marker in text

    def _ensure_metadata(self, page: DocsPage, collection_id: str) -> str:
        content = page.content_markdown or ""
        marker = f"{OUTLINE_EXTERNAL_ID_MARKER} {page.external_id}"
        if marker in content:
            return content

        meta_lines = [
            OUTLINE_METADATA_SENTINEL,
            f"{OUTLINE_EXTERNAL_ID_MARKER} {page.external_id}",
            f"collection_id: {collection_id}",
        ]
        for key, value in page.properties.items():
            meta_lines.append(f"{key}: {value}")
        if page.tags:
            meta_lines.append(f"tags: {', '.join(page.tags)}")

        metadata_block = "<details><summary>Gremlin Metadata</summary>\n" + "\n".join(meta_lines) + "\n</details>"
        if content.strip():
            return f"{content.rstrip()}\n\n{metadata_block}\n"
        return f"{metadata_block}\n"

    def _document_url(self, document: Dict[str, Any]) -> Optional[str]:
        base = self.api_base[:-4] if self.api_base.endswith("/api") else self.api_base
        url_id = document.get("urlId") or document.get("id")
        if base and url_id:
            return f"{base}/doc/{url_id}"
        return None


class ConfluenceClient(DocsSinkProvider):
    """Confluence Cloud provider implementation."""

    def __init__(
        self,
        api_base: str,
        email: str,
        api_token: str,
        default_space_key: Optional[str] = None,
        default_parent_id: Optional[str] = None,
    ) -> None:
        self.api_base = self._normalize_base(api_base)
        self.email = email
        self.api_token = api_token
        self.default_space_key = default_space_key
        self.default_parent_id = default_parent_id
        self.session = requests.Session()

    @property
    def name(self) -> str:  # pragma: no cover - trivial
        return "confluence"

    @property
    def provider_type(self) -> DocsProvider:  # pragma: no cover - trivial
        return DocsProvider.CONFLUENCE

    def _normalize_base(self, api_base: str) -> str:
        base = api_base.rstrip("/")
        if base.endswith(CONFLUENCE_DEFAULT_API_SUFFIX):
            return base
        if base.endswith("/wiki"):
            return f"{base}/rest/api"
        if base.endswith("/rest/api"):
            return base
        return f"{base}{CONFLUENCE_DEFAULT_API_SUFFIX}"

    def _headers(self) -> Dict[str, str]:
        return {
            "Content-Type": "application/json",
            "Accept": "application/json",
        }

    def _request(
        self,
        method: str,
        path: str,
        json: Optional[Dict[str, Any]] = None,
        params: Optional[Dict[str, Any]] = None,
    ) -> requests.Response:
        url = f"{self.api_base}{path}"
        resp = self.session.request(
            method,
            url,
            headers=self._headers(),
            json=json,
            params=params,
            auth=(self.email, self.api_token),
            timeout=30,
        )
        if resp.status_code == 429:
            retry_after = int(resp.headers.get("Retry-After", "1"))
            time.sleep(retry_after)
            resp = self.session.request(
                method,
                url,
                headers=self._headers(),
                json=json,
                params=params,
                auth=(self.email, self.api_token),
                timeout=30,
            )
        return resp

    def _error_message(self, resp: requests.Response) -> str:
        try:
            data = resp.json()
        except Exception:
            return f"HTTP {resp.status_code}"
        if isinstance(data, dict):
            if data.get("message"):
                return data["message"]
            if data.get("errorMessages"):
                return "; ".join(data.get("errorMessages"))
            if data.get("errors"):
                return str(data.get("errors"))
        return f"HTTP {resp.status_code}"

    def test_connection(self) -> Dict[str, Any]:
        resp = self._request("GET", "/user/current")
        if resp.status_code == 401:
            raise DocsSinkError("Confluence auth failed")
        if not resp.ok:
            raise DocsSinkError(f"Confluence connection failed ({resp.status_code})")
        data = resp.json()
        return {"status": "ok", "user": data.get("displayName") or data.get("accountId")}

    def get_space(self, space_key: str) -> Optional[Dict[str, Any]]:
        resp = self._request("GET", f"/space/{space_key}")
        if resp.status_code == 404:
            return None
        if not resp.ok:
            return None
        return resp.json()

    def get_content(self, content_id: str, expand: Optional[str] = None) -> Optional[Dict[str, Any]]:
        params = {"expand": expand} if expand else None
        resp = self._request("GET", f"/content/{content_id}", params=params)
        if resp.status_code == 404:
            return None
        if not resp.ok:
            return None
        return resp.json()

    def _search_cql(self, cql: str, limit: int = 1) -> List[Dict[str, Any]]:
        resp = self._request("GET", "/content/search", params={"cql": cql, "limit": limit})
        if not resp.ok:
            return []
        return resp.json().get("results", [])

    def _cql_escape(self, value: str) -> str:
        return value.replace("\\", "\\\\").replace('"', '\\"')

    def get_page_by_external_id(self, external_id: str, space_key: Optional[str] = None) -> Optional[Dict[str, Any]]:
        external_id = str(external_id)
        resolved_space = space_key or self.default_space_key
        if not resolved_space:
            return None
        cql = (
            f'space = "{self._cql_escape(resolved_space)}" '
            f'AND type = page '
            f'AND content.property[{CONFLUENCE_EXTERNAL_ID_KEY}].value = "{self._cql_escape(external_id)}"'
        )
        results = self._search_cql(cql, limit=1)
        if results:
            return results[0]

        marker = f"{CONFLUENCE_MARKER_PREFIX}{external_id}"
        cql = (
            f'space = "{self._cql_escape(resolved_space)}" '
            f'AND type = page '
            f'AND text ~ "{self._cql_escape(marker)}"'
        )
        results = self._search_cql(cql, limit=1)
        return results[0] if results else None

    def _build_storage_body(self, page: DocsPage) -> Dict[str, Any]:
        storage_value = markdown_to_storage(page.content_markdown or "", external_id=page.external_id)
        return {"storage": {"value": storage_value, "representation": "storage"}}

    def _extract_page_url(self, data: Dict[str, Any]) -> Optional[str]:
        links = data.get("_links", {})
        base = links.get("base")
        webui = links.get("webui")
        if base and webui:
            return f"{base}{webui}"
        if links.get("self"):
            return links.get("self")
        return None

    def _get_content_property(self, content_id: str, key: str) -> Optional[Dict[str, Any]]:
        resp = self._request("GET", f"/content/{content_id}/property/{key}")
        if resp.status_code == 404:
            return None
        if not resp.ok:
            return None
        return resp.json()

    def _set_content_property(self, content_id: str, key: str, value: Any) -> bool:
        existing = self._get_content_property(content_id, key)
        if existing and existing.get("value") == value:
            return True
        if existing:
            payload = {
                "id": existing.get("id"),
                "key": key,
                "value": value,
                "version": {"number": existing.get("version", {}).get("number", 1) + 1},
            }
            resp = self._request("PUT", f"/content/{content_id}/property/{key}", json=payload)
            return resp.ok
        payload = {"key": key, "value": value}
        resp = self._request("POST", f"/content/{content_id}/property", json=payload)
        return resp.ok

    def _add_labels(self, content_id: str, tags: List[str]) -> bool:
        if not tags:
            return True
        labels = []
        seen = set()
        for tag in tags:
            slug = _slugify_label(tag)
            if slug and slug not in seen:
                seen.add(slug)
                labels.append({"prefix": "global", "name": slug})
        if not labels:
            return True
        resp = self._request("POST", f"/content/{content_id}/label", json=labels)
        return resp.ok

    def _apply_metadata(self, content_id: str, page: DocsPage) -> Optional[str]:
        if not self._set_content_property(content_id, CONFLUENCE_EXTERNAL_ID_KEY, page.external_id):
            return "Failed to set content property."
        if page.tags and not self._add_labels(content_id, page.tags):
            return "Failed to apply labels."
        return None

    def upsert_page(self, page: DocsPage) -> DocsPublishResult:
        return self._upsert_page(page, allow_conflict_retry=True)

    def _upsert_page(self, page: DocsPage, allow_conflict_retry: bool) -> DocsPublishResult:
        space_key = page.properties.get("space_key") if page.properties else None
        resolved_space = space_key or self.default_space_key
        if not resolved_space:
            raise DocsSinkError("Confluence space_key is required.")

        parent_id = page.parent_id or self.default_parent_id
        existing = self.get_page_by_external_id(page.external_id, space_key=resolved_space)

        if existing:
            content_id = existing.get("id")
            content = self.get_content(content_id, expand="version")
            if not content:
                return DocsPublishResult(
                    external_id=page.external_id,
                    status="failed",
                    page_id=content_id,
                    page_url=None,
                    error="Failed to fetch current content version.",
                )
            version_number = content.get("version", {}).get("number", 1) + 1
            payload: Dict[str, Any] = {
                "id": content_id,
                "type": "page",
                "title": page.title,
                "space": {"key": resolved_space},
                "version": {"number": version_number},
                "body": self._build_storage_body(page),
            }
            if parent_id:
                payload["ancestors"] = [{"id": parent_id}]
            resp = self._request("PUT", f"/content/{content_id}", json=payload)
            if not resp.ok:
                return DocsPublishResult(
                    external_id=page.external_id,
                    status="failed",
                    page_id=content_id,
                    page_url=self._extract_page_url(existing),
                    error=self._error_message(resp),
                )
            data = resp.json()
            metadata_error = self._apply_metadata(content_id, page)
            if metadata_error:
                return DocsPublishResult(
                    external_id=page.external_id,
                    status="partial",
                    page_id=content_id,
                    page_url=self._extract_page_url(data),
                    error=metadata_error,
                )
            return DocsPublishResult(
                external_id=page.external_id,
                status="updated",
                page_id=content_id,
                page_url=self._extract_page_url(data),
            )

        payload = {
            "type": "page",
            "title": page.title,
            "space": {"key": resolved_space},
            "body": self._build_storage_body(page),
        }
        if parent_id:
            payload["ancestors"] = [{"id": parent_id}]
        resp = self._request("POST", "/content", json=payload)
        if resp.status_code == 409:
            if not allow_conflict_retry:
                return DocsPublishResult(
                    external_id=page.external_id,
                    status="failed",
                    page_id=None,
                    page_url=None,
                    error="Conflict creating page.",
                )
            existing = self.get_page_by_external_id(page.external_id, space_key=resolved_space)
            if existing:
                return self._upsert_page(
                    DocsPage(
                        external_id=page.external_id,
                        title=page.title,
                        content_markdown=page.content_markdown,
                        parent_id=parent_id,
                        properties={"space_key": resolved_space},
                        tags=page.tags,
                    ),
                    allow_conflict_retry=False,
                )
        if not resp.ok:
            return DocsPublishResult(
                external_id=page.external_id,
                status="failed",
                page_id=None,
                page_url=None,
                error=self._error_message(resp),
            )
        data = resp.json()
        content_id = data.get("id")
        if content_id:
            metadata_error = self._apply_metadata(content_id, page)
            if metadata_error:
                return DocsPublishResult(
                    external_id=page.external_id,
                    status="partial",
                    page_id=content_id,
                    page_url=self._extract_page_url(data),
                    error=metadata_error,
                )
        return DocsPublishResult(
            external_id=page.external_id,
            status="created",
            page_id=content_id,
            page_url=self._extract_page_url(data),
        )

    def upsert_pages_batch(self, pages: List[DocsPage]) -> BatchOutcome:
        outcome = BatchOutcome()
        for page in pages:
            result = self.upsert_page(page)
            if result.status in ("created", "updated"):
                outcome.succeeded.append(result.to_dict())
            elif result.status == "failed":
                outcome.failed.append(result.to_dict())
            else:
                outcome.retryable_failed.append(result.to_dict())
        return outcome

    def delete_page(self, page_id: str) -> bool:
        resp = self._request("DELETE", f"/content/{page_id}")
        return resp.ok


__all__ = [
    "DocsProvider",
    "DocsPage",
    "DocsPublishResult",
    "DocsPublishLogEntry",
    "DocsSinkProvider",
    "DocsSinkError",
    "NotionClient",
    "OutlineClient",
    "ConfluenceClient",
]
